from collections import UserList, defaultdict

from pydantic import BaseModel


class Issue(BaseModel):
    """Base class for all issues"""

    message: str
    code: str | None = None
    fix: str | None = None

    @classmethod
    def issue_type(cls) -> str:
        return cls.__name__


class ModelSyntaxError(Issue):
    """If any syntax error is found. Stop validation
    and ask user to fix the syntax error first."""

    ...


class ImplementationWarning(Issue):
    """This is only for conceptual data model. It means that conversion to DMS
    will fail unless user implements the missing part."""

    ...


class ConsistencyError(Issue):
    """If any consistency error is found, the deployment of the data model will fail. For example,
    if a reverse direct relations points to a non-existing direct relation. This is only relevant for
    DMS model.
    """

    ...


class Recommendation(Issue):
    """Best practice recommendation."""

    ...


class IssueList(UserList[Issue]):
    """A list of issues that can be sorted by type and message."""

    def by_type(self) -> dict[type[Issue], list[Issue]]:
        """Returns a dictionary of issues sorted by their type."""
        result: dict[type[Issue], list[Issue]] = defaultdict(list)
        for issue in self.data:
            issue_type = type(issue)
            if issue_type not in result:
                result[issue_type] = []
            result[issue_type].append(issue)
        return result

    def by_code(self) -> dict[str, list[Issue]]:
        """Returns a dictionary of issues sorted by their code."""
        result: dict[str, list[Issue]] = defaultdict(list)
        for issue in self.data:
            if issue.code is not None:
                result[issue.code].append(issue)
            else:
                result["UNDEFINED"].append(issue)
        return dict(result)



from collections.abc import Iterator, Sequence
from typing import TypeVar

T_Sequence = TypeVar("T_Sequence", bound=Sequence)


def chunker_sequence(sequence: T_Sequence, size: int) -> Iterator[T_Sequence]:
    """Yield successive n-sized chunks from sequence."""
    for i in range(0, len(sequence), size):
        # MyPy does not expect sequence[i : i + size] to be of type T_Sequence
        yield sequence[i : i + size]  # type: ignore[misc]



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from cognite.neat._utils.useful_types import BaseModelObject, T_Reference


class Resource(BaseModelObject):
    """Base class for all data modeling resources."""

    ...


T_Resource = TypeVar("T_Resource", bound=Resource)


class WriteableResource(Resource, Generic[T_Resource], ABC):
    """Base class for all writeable data modeling resources."""

    @abstractmethod
    def as_request(self) -> T_Resource:
        """Convert the response model to a request model by removing read-only fields."""
        raise NotImplementedError()


T_Response = TypeVar("T_Response", bound=WriteableResource)


class APIResource(Generic[T_Reference], ABC):
    """Base class for all API data modeling resources."""

    @abstractmethod
    def as_reference(self) -> T_Reference:
        """Convert the resource to a reference object (identifier)."""
        raise NotImplementedError()



from abc import ABC
from typing import Annotated, Any, Literal, TypeAlias, get_args

from pydantic import BeforeValidator, Field, JsonValue, TypeAdapter, model_serializer
from pydantic_core.core_schema import FieldSerializationInfo

from cognite.neat._utils.text import humanize_collection
from cognite.neat._utils.useful_types import BaseModelObject

from ._references import ContainerReference, NodeReference, ViewReference

# Base classes and helpers


class Parameter(BaseModelObject):
    parameter: str


class FilterDataDefinition(BaseModelObject, ABC):
    """Base class for filter data models."""

    # This is an internal field used for discriminating between filter types. It is not part of the actual
    # data sent to or received from the API. See the _move_filter_key function for more details.
    filter_type: str = Field(..., exclude=True)


class PropertyReference(FilterDataDefinition, ABC):
    """Represents the property path in filters."""

    property: list[str] = Field(..., min_length=2, max_length=3)


## Leaf filters that follows the standard pattern


class EqualsFilterData(PropertyReference):
    filter_type: Literal["equals"] = Field("equals", exclude=True)
    value: JsonValue | PropertyReference


class InFilterData(PropertyReference):
    filter_type: Literal["in"] = Field("in", exclude=True)
    values: list[JsonValue] | PropertyReference


class RangeFilterData(PropertyReference):
    filter_type: Literal["range"] = Field("range", exclude=True)
    gt: str | int | float | PropertyReference | None = None
    gte: str | int | float | PropertyReference | None = None
    lt: str | int | float | PropertyReference | None = None
    lte: str | int | float | PropertyReference | None = None


class PrefixFilterData(PropertyReference):
    filter_type: Literal["prefix"] = Field("prefix", exclude=True)
    value: str | Parameter


class ExistsFilterData(PropertyReference):
    filter_type: Literal["exists"] = Field("exists", exclude=True)


class ContainsAnyFilterData(PropertyReference):
    filter_type: Literal["containsAny"] = Field("containsAny", exclude=True)
    values: list[JsonValue] | PropertyReference


class ContainsAllFilterData(PropertyReference):
    filter_type: Literal["containsAll"] = Field("containsAll", exclude=True)
    values: list[JsonValue] | PropertyReference


class MatchAllFilterData(FilterDataDefinition):
    filter_type: Literal["matchAll"] = Field("matchAll", exclude=True)


class NestedFilterData(FilterDataDefinition):
    filter_type: Literal["nested"] = Field("nested", exclude=True)
    scope: list[str] = Field(..., min_length=1, max_length=3)
    filter: "Filter"


class OverlapsFilterData(PropertyReference):
    filter_type: Literal["overlaps"] = Field("overlaps", exclude=True)
    start_property: list[str] = Field(..., min_length=1, max_length=3)
    end_property: list[str] = Field(..., min_length=1, max_length=3)
    gt: str | int | float | PropertyReference | None = None
    gte: str | int | float | PropertyReference | None = None
    lt: str | int | float | PropertyReference | None = None
    lte: str | int | float | PropertyReference | None = None


class ListFilterDataDefinition(FilterDataDefinition, ABC):
    """Base class for filters that operate on lists of values."""

    data: list[Any]


## Leaf filters with custom serialization logic due to creativity in the API design


class HasDataFilter(ListFilterDataDefinition):
    filter_type: Literal["hasData"] = Field("hasData", exclude=True)
    data: list[ViewReference | ContainerReference]

    # MyPy complains about thet signature of the method here, even though its compatible with the pydantic source code.
    # And tests are passing fine.
    @model_serializer(mode="plain")  # type: ignore[type-var]
    def serialize_model(self, info: FieldSerializationInfo) -> list[dict[str, Any]]:
        output: list[dict[str, Any]] = []
        for item in self.data:
            item_dict = item.model_dump(**vars(info))
            if isinstance(item, ViewReference):
                item_dict["type"] = "view"
            elif isinstance(item, ContainerReference):
                item_dict["type"] = "container"
            output.append(item_dict)
        return output


class InstanceReferencesFilterData(ListFilterDataDefinition):
    filter_type: Literal["instanceReferences"] = Field("instanceReferences", exclude=True)
    data: list[NodeReference]

    # MyPy complains about thet signature of the method here, even though its compatible with the pydantic source code.
    # And tests are passing fine.
    @model_serializer(mode="plain")  # type: ignore[type-var]
    def serialize_model(self, info: FieldSerializationInfo) -> list[dict[str, Any]]:
        return [item.model_dump(**vars(info)) for item in self.data]


## Logical filters combining other filters


class AndFilter(ListFilterDataDefinition):
    filter_type: Literal["and"] = Field("and", exclude=True)
    data: "list[Filter]"

    # MyPy complains about thet signature of the method here, even though its compatible with the pydantic source code.
    # And tests are passing fine.
    @model_serializer(mode="plain")  # type: ignore[type-var]
    def serialize_model(self, info: FieldSerializationInfo) -> list[dict[str, Any]]:
        return [FilterAdapter.dump_python(item, **vars(info)) for item in self.data]


class OrFilter(ListFilterDataDefinition):
    filter_type: Literal["or"] = Field("or", exclude=True)
    data: "list[Filter]"

    # MyPy complains about thet signature of the method here, even though its compatible with the pydantic source code.
    # And tests are passing fine.
    @model_serializer(mode="plain")  # type: ignore[type-var]
    def serialize_model(self, info: FieldSerializationInfo) -> list[dict[str, Any]]:
        return [FilterAdapter.dump_python(item, **vars(info)) for item in self.data]


class NotFilter(FilterDataDefinition):
    filter_type: Literal["not"] = Field("not", exclude=True)
    data: "Filter"

    # MyPy complains about thet signature of the method here, even though its compatible with the pydantic source code.
    # And tests are passing fine.
    @model_serializer(mode="plain")  # type: ignore[type-var]
    def serialize_model(self, info: FieldSerializationInfo) -> dict[str, Any]:
        return FilterAdapter.dump_python(self.data, **vars(info))


FilterData = Annotated[
    EqualsFilterData
    | PrefixFilterData
    | InFilterData
    | RangeFilterData
    | ExistsFilterData
    | ContainsAnyFilterData
    | ContainsAllFilterData
    | MatchAllFilterData
    | NestedFilterData
    | OverlapsFilterData
    | HasDataFilter
    | InstanceReferencesFilterData
    | AndFilter
    | OrFilter
    | NotFilter,
    Field(discriminator="filter_type"),
]


FilterTypes: TypeAlias = Literal[
    "equals",
    "prefix",
    "in",
    "range",
    "exists",
    "containsAny",
    "containsAll",
    "matchAll",
    "nested",
    "overlaps",
    "hasData",
    "instanceReferences",
    "and",
    "or",
    "not",
]

LegacyFilterTypes: TypeAlias = Literal["invalid"]

AVAILABLE_FILTERS: frozenset[str] = frozenset(get_args(FilterTypes))
LEGACY_FILTERS: frozenset[str] = frozenset(get_args(LegacyFilterTypes))


def _move_filter_key(value: Any) -> Any:
    """The DMS API filters have an unusual structure.

    It has the filter type as the key of the outer dict, and then the actual filter data as the value, e.g.,
    {
        "equals": {
            "property": [...],
            "value": ...
        }
    }
    We could have modeled it that way with Pydantic, and had an union of pydantic models of all possible filter types.
    However, validating union types in Pydantic without a discriminator leads to poor error messages. If the filter
    data does not comply with any of the union types, Pydantic will give one error message per union type. For exampl,
    if the user writes
    {
        "equals": {
            "property": "my_property"  # Should be a list,
            "value": 'my_value'
            }
    }
    Pydantic will give 15 error messages, one for each filter type in the union, saying that the data does not
    comply with that filter type. This is not very user-friendly.

    Instead, we introduce an internal field "filter_type" inside the filter data models, and use that as a
    discriminator. This will enable the validation to be two steps. First, we validate the outer key and
    that it is a known filter type. Then, we move that key inside the filter data as the "filter_type" field, and
    validate the filter data against the correct model based on that discriminator.


    This function transforms the data from the outer-key format to the inner-key format. For example, it transforms
    the equals filter form above into

    {
        "equals": {
            "property": [...],
            "value": ...,
            "filterType": "equals"
        }
    }
    """
    # legacy filter which we want to ignore
    if _is_legacy_filter(value):
        return None
    if not isinstance(value, dict):
        return value
    if len(value) != 1:
        raise ValueError("Filter data must have exactly one key.")
    if "filterType" in value:
        # Already in the correct format
        return value
    key, data = next(iter(value.items()))
    # Check if inner data already has filterType (already processed by a previous recursive call)
    if isinstance(data, dict) and "filterType" in data:
        return value
    if key not in AVAILABLE_FILTERS:
        raise ValueError(
            f"Unknown filter type: {key!r}. Available filter types: {humanize_collection(AVAILABLE_FILTERS)}."
        )
    if isinstance(data, dict) and key == "not":
        # Not is a recursive filter, so we need to move the filter key inside its data as well
        output = _move_filter_key(data.copy())
        return {key: {"filterType": key, "data": output}}
    elif isinstance(data, dict):
        output = data.copy()
        output["filterType"] = key
        return {key: output}
    elif isinstance(data, list) and key in {"and", "or"}:
        # And and Or are recursive filters, so we need to move the filter key inside each of their data items as well
        return {key: {"filterType": key, "data": [_move_filter_key(item) for item in data]}}
    elif isinstance(data, list):
        # Leaf list filters, hasData and instanceReferences
        return {key: {"filterType": key, "data": data}}
    else:
        # Let the regular validation handle it (possible not an issue)
        return value


def _is_legacy_filter(filter_obj: Any) -> bool:
    """Check if filter is a legacy filter no longer supported by DMS API"""

    def traverse(obj: Any) -> bool:
        if isinstance(obj, dict):
            for key, value in obj.items():
                if key in LEGACY_FILTERS:
                    return True
                if traverse(value):
                    return True
        elif isinstance(obj, list):
            for item in obj:
                if traverse(item):
                    return True
        return False

    return traverse(filter_obj)


Filter = Annotated[dict[FilterTypes, FilterData] | None, BeforeValidator(_move_filter_key)]

FilterAdapter: TypeAdapter[Filter] = TypeAdapter(Filter)



from abc import ABC, abstractmethod
from typing import Any


class NeatPlugin(ABC):
    @abstractmethod
    def configure(self, *args: Any, **kwargs: Any) -> Any:
        """A method that all plugins must implement."""
        raise NotImplementedError()



from collections.abc import Hashable
from datetime import date, datetime, time, timedelta
from typing import Literal, TypeAlias, TypeVar

from pydantic import BaseModel
from pydantic.alias_generators import to_camel

JsonVal: TypeAlias = None | str | int | float | bool | dict[str, "JsonVal"] | list["JsonVal"]
PrimaryTypes: TypeAlias = str | int | float | bool

T_ID = TypeVar("T_ID", bound=Hashable)
# These are the types that openpyxl supports in cells
CellValueType: TypeAlias = str | int | float | bool | datetime | date | time | timedelta | None

# The format expected for excel sheets representing a data model
DataModelTableType: TypeAlias = dict[str, list[dict[str, CellValueType]]]
PrimitiveType: TypeAlias = str | int | float | bool


class BaseModelObject(BaseModel, alias_generator=to_camel, extra="ignore"):
    """Base class for all object. This includes resources and nested objects."""

    ...


T_Item = TypeVar("T_Item", bound=BaseModelObject)


class ReferenceObject(BaseModelObject, frozen=True, populate_by_name=True):
    """Base class for all reference objects - these are identifiers."""

    ...


T_Reference = TypeVar("T_Reference", bound=ReferenceObject, covariant=True)

ModusOperandi: TypeAlias = Literal["rebuild", "additive"]



"""Validators for checking containers in the data model."""

from cognite.neat._data_model.models.dms._view_property import ViewCorePropertyRequest
from cognite.neat._data_model.rules.dms._base import DataModelRule
from cognite.neat._issues import ConsistencyError

BASE_CODE = "NEAT-DMS-CONTAINER"


class ExternalContainerDoesNotExist(DataModelRule):
    """
    Validates that any container referenced by a view property, when the
    referenced container does not belong to the data model's space, exists in CDF.

    ## What it does
    For each view property that maps to a container in a different space than the data model,
    this validator checks that the referenced external container exists in CDF.

    ## Why is this bad?
    If a view property references a container that does not exist in CDF,
    the data model cannot be deployed. The affected view property will not function, and the
    deployment of the entire data model will fail.

    ## Example
    View `my_space:WindTurbine` has a property `location` that maps to container
    `other_space:WindTurbineContainer`, where `other_space` differs from `my_space`. If that
    container does not exist in CDF, the model cannot be deployed.
    """

    code = f"{BASE_CODE}-001"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        if not self.validation_resources.merged_data_model.views:
            return errors

        for view_ref in self.validation_resources.merged_data_model.views:
            view = self.validation_resources.select_view(view_ref)

            if not view:
                raise RuntimeError(
                    f"{type(self).__name__}: View {view_ref!s} not found in local resources. This is a bug in NEAT."
                )

            if view.properties is None:
                continue

            for property_ref, property_ in view.properties.items():
                if not isinstance(property_, ViewCorePropertyRequest):
                    continue

                if property_.container.space == self.validation_resources.merged_data_model.space:
                    continue

                # Check existence of container in CDF
                if property_.container not in self.validation_resources.cdf.containers:
                    errors.append(
                        ConsistencyError(
                            message=(
                                f"View {view_ref!s} property {property_ref!s} maps to "
                                f"external container {property_.container!s} which does not exist in CDF."
                            ),
                            fix="Define necessary container in CDF",
                            code=self.code,
                        )
                    )

        return errors


class ExternalContainerPropertyDoesNotExist(DataModelRule):
    """
    Validates that any container property referenced by a view property, when the
    referenced container does not belong to the data model's space, exists in CDF.

    ## What it does
    For each view property that maps to a container in a different space than the data model,
    this validator checks that the referenced container property exists in that external container in CDF.
    This validator only runs if the external container exists in CDF.

    ## Why is this bad?
    If a view property references a container property that does not exist in CDF,
    the data model cannot be deployed. The affected view property will not function, and the
    deployment of the entire data model will fail.

    ## Example
    View `my_space:WindTurbine` has a property `location` that maps to container property
    `gpsCoordinates` in `other_space:WindTurbineContainer`. If `gpsCoordinates` does not exist
    in that container in CDF, deployment will fail.
    """

    code = f"{BASE_CODE}-002"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        if self.validation_resources.merged_data_model.views:
            for view_ref in self.validation_resources.merged_data_model.views:
                view = self.validation_resources.select_view(view_ref)

                if not view:
                    raise RuntimeError(
                        f"{type(self).__name__}: View {view_ref!s} not found in local resources. This is a bug in NEAT."
                    )

                if view.properties is None:
                    continue

                for property_ref, property_ in view.properties.items():
                    if not isinstance(property_, ViewCorePropertyRequest):
                        continue

                    if property_.container.space == self.validation_resources.merged_data_model.space:
                        continue

                    # Only check property if container exists in CDF
                    # this check is done in ExternalContainerDoesNotExist
                    if property_.container not in self.validation_resources.cdf.containers:
                        continue

                    # Check existence of container property in CDF
                    if (
                        property_.container_property_identifier
                        not in self.validation_resources.cdf.containers[property_.container].properties
                    ):
                        errors.append(
                            ConsistencyError(
                                message=(
                                    f"View {view_ref!s} property {property_ref!s} maps to "
                                    f"external container {property_.container!s} which does not have "
                                    f"property '{property_.container_property_identifier}' in CDF."
                                ),
                                fix="Define necessary container property in CDF",
                                code=self.code,
                            )
                        )

        return errors


class RequiredContainerDoesNotExist(DataModelRule):
    """
    Validates that any container required by another container exists in the data model.

    ## What it does
    For each container in the data model, this validator checks that any container it
    requires (via requires constraints) exists either in the data model or in CDF.

    ## Why is this bad?
    If a container requires another container that does not exist in the data model or in CDF,
    the data model cannot be deployed. The affected container will not function, and
    the deployment of the entire data model will fail.

    ## Example
    Container `windy_space:WindTurbineContainer` has a constraint requiring `windy_space:LocationContainer`.
    If `windy_space:LocationContainer` does not exist in the data model or in CDF, deployment will fail.
    """

    code = f"{BASE_CODE}-003"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        for container_ref in self.validation_resources.merged.containers:
            container = self.validation_resources.select_container(container_ref)

            if not container:
                raise RuntimeError(
                    f"{type(self).__name__}: Container {container_ref!s} "
                    "not found in local resources. This is a bug in NEAT."
                )

            for constraint_ref, constraint in self.validation_resources.get_requires_constraints(container):
                if not self.validation_resources.select_container(constraint.require):
                    errors.append(
                        ConsistencyError(
                            message=(
                                f"Container '{container_ref!s}' constraint '{constraint_ref}' requires container "
                                f"'{constraint.require!s}' which does not exist."
                            ),
                            fix="Define necessary container in the data model",
                            code=self.code,
                        )
                    )

        return errors


class RequiresConstraintCycle(DataModelRule):
    """
    Validates that requires constraints between containers do not form cycles.

    ## What it does
    This validator checks if the requires constraints between containers form a cycle.
    For example, if container A requires B, B requires C, and C requires A, this forms
    a cycle.

    ## Why is this bad?
    Cycles in requires constraints will be rejected by the CDF API. The deployment
    of the data model will fail if any such cycle exists.

    ## Example
    Container `my_space:OrderContainer` requires `my_space:CustomerContainer`, which
    requires `my_space:OrderContainer`. This creates a cycle and will be rejected.
    """

    code = f"{BASE_CODE}-005"
    issue_type = ConsistencyError
    alpha = True  # Still in development

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []
        optimal_edges = self.validation_resources.oriented_mst_edges

        for cycle in self.validation_resources.requires_constraint_cycles:
            cycle_str = " -> ".join(str(c) for c in cycle) + f" -> {cycle[0]!s}"

            # Find edges in cycle that are NOT in optimal structure (these should be removed)
            edges_to_remove = []
            for i, container in enumerate(cycle):
                next_container = cycle[(i + 1) % len(cycle)]
                edge = (container, next_container)
                if edge not in optimal_edges:
                    edges_to_remove.append(f"{container} -> {next_container}")

            message = f"Requires constraints form a cycle: {cycle_str}"
            if edges_to_remove:
                message += f". Recommended removal: {', '.join(edges_to_remove)} (not in optimal structure)"

            errors.append(
                ConsistencyError(
                    message=message,
                    fix="Remove one of the requires constraints to break the cycle",
                    code=self.code,
                )
            )

        return errors



from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from functools import partial
from typing import cast

from cognite.neat._client import NeatClient
from cognite.neat._data_model._constants import COGNITE_SPACES
from cognite.neat._data_model._shared import OnSuccessResultProducer
from cognite.neat._data_model._snapshot import SchemaSnapshot
from cognite.neat._data_model.models.dms import (
    ContainerRequest,
    DataModelBody,
    RequestSchema,
    T_DataModelResource,
    T_ResourceId,
)
from cognite.neat._utils.collection import chunker_sequence
from cognite.neat._utils.http_client import (
    FailedRequestItems,
    FailedResponseItems,
    ItemIDBody,
    ItemsRequest,
    SuccessResponseItems,
)
from cognite.neat._utils.http_client._data_classes import APIResponse
from cognite.neat._utils.useful_types import ModusOperandi, T_Reference

from ._differ import ItemDiffer
from ._differ_container import ContainerDiffer
from ._differ_data_model import DataModelDiffer
from ._differ_space import SpaceDiffer
from ._differ_view import ViewDiffer
from .data_classes import (
    AddedField,
    AppliedChanges,
    ChangedFieldResult,
    ContainerDeploymentPlan,
    DataModelEndpoint,
    DeploymentResult,
    FieldChange,
    FieldChanges,
    HTTPChangeResult,
    NoOpChangeResult,
    RemovedField,
    ResourceChange,
    ResourceDeploymentPlan,
    ResourceDeploymentPlanList,
    SeverityType,
)


@dataclass
class DeploymentOptions:
    """Configuration options for deployment.

    Attributes:
        dry_run (bool): If True, the deployment will be simulated without applying changes. Defaults to True.
        auto_rollback (bool): If True, automatically roll back changes if deployment fails. Defaults to True.
        max_severity (SeverityType): Maximum allowed severity of changes to proceed with deployment.
            Defaults to SeverityType.SAFE.
        modus_operandi (ModusOperandi): Deployment mode. If "additive", only add/modify resources
            specified in the data model. If "rebuild", remove resources not present in the data model.
            Defaults to "additive".
    """

    dry_run: bool = True
    auto_rollback: bool = True
    drop_data: bool = False
    max_severity: SeverityType = SeverityType.WARNING
    modus_operandi: ModusOperandi = "additive"


class SchemaDeployer(OnSuccessResultProducer):
    INDEX_DELETE_BATCH_SIZE = 10
    CONSTRAINT_DELETE_BATCH_SIZE = 10

    def __init__(self, client: NeatClient, options: DeploymentOptions | None = None) -> None:
        super().__init__()
        self.client: NeatClient = client
        self.options: DeploymentOptions = options or DeploymentOptions()

    def run(self, data_model: RequestSchema) -> None:
        if self._results is not None:
            raise RuntimeError("SchemaDeployer has already been run.")
        self._results = self.deploy(data_model)

    def deploy(self, data_model: RequestSchema) -> DeploymentResult:
        # Step 1: Fetch current CDF state
        snapshot = SchemaSnapshot.fetch_cdf_data_model(self.client, data_model)

        # Step 2: Create deployment plan by comparing local vs cdf
        plan = self.create_deployment_plan(snapshot, data_model)

        # Step 3: Adjust plan based on modus operandi
        if self.options.modus_operandi == "additive":
            # Filter out deletions and removals in additive mode
            plan = plan.consolidate_changes()
        elif self.options.modus_operandi == "rebuild":
            # Breaking changes are forced by deleting and recreating resources
            # Containers are treated as additive unless drop_data is specified
            plan = plan.force_changes(self.options.drop_data)
        else:
            raise NotImplementedError(f"Unsupported modus operandi: {self.options.modus_operandi!r}")

        if not self.should_proceed_to_deploy(plan):
            # Step 4: Check if deployment should proceed
            return DeploymentResult(status="failure", plan=list(plan), snapshot=snapshot)
        elif self.options.dry_run:
            # Step 5: If dry-run, return plan without applying
            return DeploymentResult(status="pending", plan=list(plan), snapshot=snapshot)

        # Step 6: Apply changes
        changes = self.apply_changes(plan)

        # Step 7: Rollback if failed and auto_rollback is enabled
        if not changes.is_success and self.options.auto_rollback:
            recovery = self.apply_changes(changes.as_recovery_plan())
            return DeploymentResult(
                status="recovered" if recovery.is_success else "recovery_failed",
                plan=list(plan),
                snapshot=snapshot,
                responses=changes,
                recovery=recovery,
            )
        return DeploymentResult(
            status="success" if changes.is_success else "partial", plan=list(plan), snapshot=snapshot, responses=changes
        )

    def create_deployment_plan(self, snapshot: SchemaSnapshot, data_model: RequestSchema) -> ResourceDeploymentPlanList:
        return ResourceDeploymentPlanList(
            [
                self._create_resource_plan(
                    snapshot.spaces,
                    data_model.spaces,
                    "spaces",
                    SpaceDiffer(),
                    skip_criteria=partial(self._skip_resource, model_space=data_model.data_model.space),
                ),
                self._create_resource_plan(
                    snapshot.containers,
                    data_model.containers,
                    "containers",
                    ContainerDiffer(),
                    ContainerDeploymentPlan,
                    skip_criteria=partial(self._skip_resource, model_space=data_model.data_model.space),
                ),
                self._create_resource_plan(
                    snapshot.views,
                    data_model.views,
                    "views",
                    ViewDiffer(
                        current_container_map=snapshot.containers,
                        new_container_map={container.as_reference(): container for container in data_model.containers},
                    ),
                    skip_criteria=partial(self._skip_resource, model_space=data_model.data_model.space),
                ),
                self._create_resource_plan(
                    snapshot.data_model,
                    [data_model.data_model],
                    "datamodels",
                    DataModelDiffer(),
                    skip_criteria=partial(self._skip_resource, model_space=data_model.data_model.space),
                ),
            ]
        )

    def _create_resource_plan(
        self,
        current_resources: dict[T_ResourceId, T_DataModelResource],
        new_resources: list[T_DataModelResource],
        endpoint: DataModelEndpoint,
        differ: ItemDiffer[T_DataModelResource],
        plan_type: type[ResourceDeploymentPlan[T_ResourceId, T_DataModelResource]] = ResourceDeploymentPlan,
        skip_criteria: Callable[[T_ResourceId], str | None] | None = None,
    ) -> ResourceDeploymentPlan[T_ResourceId, T_DataModelResource]:
        resources: list[ResourceChange[T_ResourceId, T_DataModelResource]] = []
        for new_resource in new_resources:
            # We know that .as_reference() will return T_ResourceId
            ref = cast(T_ResourceId, new_resource.as_reference())
            if skip_criteria is not None and (reason := skip_criteria(ref)):
                resources.append(ResourceChange(resource_id=ref, new_value=None, current_value=None, message=reason))
                continue
            if ref not in current_resources:
                resources.append(ResourceChange(resource_id=ref, new_value=new_resource))
                continue
            current_resource = current_resources[ref]
            diffs = differ.diff(current_resource, new_resource)
            if isinstance(current_resource, ContainerRequest) and isinstance(new_resource, ContainerRequest):
                # CDF doesn't support in-place modification of constraints/indexes,
                # so we transform changes to remove + add operations in both modes
                diffs = self.remove_readd_modified_indexes_and_constraints(diffs, current_resource, new_resource)

            warnings = self._generate_warnings_for_constraint_and_index_changes(diffs)
            resources.append(
                ResourceChange(
                    resource_id=ref,
                    new_value=new_resource,
                    current_value=current_resource,
                    changes=diffs,
                    message=" ".join(warnings) if warnings else None,
                )
            )

        return plan_type(endpoint=endpoint, resources=resources)

    @classmethod
    def remove_readd_modified_indexes_and_constraints(
        cls, diffs: list[FieldChange], current_resource: ContainerRequest, new_resource: ContainerRequest
    ) -> list[FieldChange]:
        """Constraints and indexes cannot be modified directly; they must be removed and re-added.

        Args:
            diffs: The list of field changes detected by the differ.
            current_resource: The current state of the container.
            new_resource: The desired state of the container.
        Returns:
            A modified list of field changes with constraints and indexes handled appropriately.
        """
        modified_diffs: list[FieldChange] = []
        for diff in diffs:
            if (diff.field_path.startswith("constraints") or diff.field_path.startswith("indexes")) and isinstance(
                diff, FieldChanges
            ):
                if "." not in diff.field_path:
                    # Should not happen, but just in case
                    raise RuntimeError("Bug in Neat. Malformed field path for constraint/index change.")
                # Field type is either "constraints" or "indexes"
                field_type, identifier, *_ = diff.field_path.split(".", maxsplit=2)
                field_path = f"{field_type}.{identifier}"
                modified_diffs.append(
                    RemovedField(
                        field_path=field_path,
                        item_severity=SeverityType.WARNING,
                        current_value=getattr(current_resource, field_type)[identifier],
                    )
                )
                add_severity = SeverityType.WARNING if field_type == "constraints" else SeverityType.SAFE
                modified_diffs.append(
                    AddedField(
                        field_path=field_path,
                        item_severity=add_severity,
                        new_value=getattr(new_resource, field_type)[identifier],
                    )
                )
            else:
                modified_diffs.append(diff)
        return modified_diffs

    @classmethod
    def _skip_resource(cls, resource_id: T_ResourceId, model_space: str) -> str | None:
        """Checks if a resource should be skipped based on its space.

        Args:
            resource_id: The ID of the resource to check.
            model_space: The space of the data model.

        Returns:
            A reason for skipping if the resource space does not match the model space, otherwise None.
        """
        if resource_id.space in COGNITE_SPACES:
            return f"Skipping resource as it is in the reserved Cognite space '{resource_id.space}'."
        elif resource_id.space != model_space:
            return (
                f"Skipping resource as it is in the space '{resource_id.space}'"
                f" not matching data model space '{model_space}'."
            )
        return None

    @classmethod
    def _generate_warnings_for_constraint_and_index_changes(cls, diffs: list[FieldChange]) -> list[str]:
        """Generate warning messages for constraint and index changes.

        Args:
            diffs: The list of field changes.

        Returns:
            A list of warning messages for field changes involving constraint and index changes.
        """
        warnings: list[str] = []
        if any(isinstance(diff, AddedField) and diff.field_path.startswith("constraints.") for diff in diffs):
            warnings.append(
                "Adding constraints could cause ingestion failures if the data being ingested violates the constraint."
            )
        if any(
            isinstance(diff, RemovedField)
            and (diff.field_path.startswith("constraints.") or diff.field_path.startswith("indexes."))
            for diff in diffs
        ):
            warnings.append("Removing constraints or indexes may affect query performance.")
        return warnings

    def should_proceed_to_deploy(self, plan: Sequence[ResourceDeploymentPlan]) -> bool:
        max_severity_in_plan = SeverityType.max_severity(
            [change.severity for resource_plan in plan for change in resource_plan.resources],
            default=SeverityType.SAFE,
        )
        return max_severity_in_plan.value <= self.options.max_severity.value

    def apply_changes(self, plan: Sequence[ResourceDeploymentPlan]) -> AppliedChanges:
        """Applies the given deployment plan to CDF by making the necessary API calls.

        Args:
            plan (list[ResourceDeploymentPlan]): The deployment plan to apply.

        Returns:
            AppliedChanges: The result of applying the changes.
        """
        applied_changes = AppliedChanges()
        # If any HTTP request fails, the skip_message will be set and subsequent operations will be skipped
        failure_message: str | None = None
        for resource in reversed(plan):
            if failure_message is None:
                deletions = self._delete_items(resource)
                applied_changes.deletions.extend(deletions)
                if any(not deletion.is_success for deletion in deletions):
                    failure_message = f"Skipping due to {resource.endpoint} deletions failing."
            else:
                applied_changes.skipped.extend(
                    [
                        NoOpChangeResult(endpoint=resource.endpoint, change=change, reason=failure_message)
                        for change in resource.to_delete
                    ]
                )

        for resource in plan:
            if failure_message is None:
                if isinstance(resource, ContainerDeploymentPlan):
                    # Note that we continue to deploy even if removing constraints/indexes fail,
                    # as the creation/update of views and data models will still succeed.
                    applied_changes.changed_fields.extend(self._remove_container_constraints(resource))
                    applied_changes.changed_fields.extend(self._remove_container_indexes(resource))

                creations, updated = self._upsert_items(resource)
                applied_changes.created.extend(creations)
                applied_changes.updated.extend(updated)
                if any(not change.is_success for change in creations + updated):
                    failure_message = f"Skipping due to {resource.endpoint} upsert failing."
            else:
                applied_changes.skipped.extend(
                    [
                        NoOpChangeResult(endpoint=resource.endpoint, change=change, reason=failure_message)
                        for change in resource.to_upsert
                    ]
                )

            applied_changes.unchanged.extend(
                [
                    NoOpChangeResult(endpoint=resource.endpoint, change=change, reason="No changes detected.")
                    for change in resource.unchanged
                ]
            )
            applied_changes.skipped.extend(
                [
                    NoOpChangeResult(endpoint=resource.endpoint, change=change, reason=change.message or "Unknown")
                    for change in resource.skip
                ]
            )
        return applied_changes

    def _delete_items(self, resource: ResourceDeploymentPlan) -> list[HTTPChangeResult]:
        to_delete_by_id = {change.resource_id: change for change in resource.to_delete}
        if not to_delete_by_id:
            return []
        responses = self.client.http_client.request_with_retries(
            ItemsRequest(
                endpoint_url=self.client.config.create_api_url(f"/models/{resource.endpoint}/delete"),
                method="POST",
                body=ItemIDBody(items=list(to_delete_by_id.keys())),
            )
        )
        return self._process_resource_responses(responses, to_delete_by_id, resource.endpoint)

    def _upsert_items(self, resource: ResourceDeploymentPlan) -> tuple[list[HTTPChangeResult], list[HTTPChangeResult]]:
        to_upsert = [
            resource_change.new_value for resource_change in resource.to_upsert if resource_change.new_value is not None
        ]
        if not to_upsert:
            return [], []
        responses = self.client.http_client.request_with_retries(
            ItemsRequest(
                endpoint_url=self.client.config.create_api_url(f"/models/{resource.endpoint}"),
                method="POST",
                body=DataModelBody(items=to_upsert),
            )
        )
        to_create_by_id = {rc.resource_id: rc for rc in resource.to_create}
        create_result = self._process_resource_responses(responses, to_create_by_id, resource.endpoint)
        to_update_by_id = {rc.resource_id: rc for rc in resource.to_update}
        update_result = self._process_resource_responses(responses, to_update_by_id, resource.endpoint)
        return create_result, update_result

    def _remove_container_indexes(self, resource: ContainerDeploymentPlan) -> list[ChangedFieldResult]:
        return self._remove_container_fields(
            resource.indexes_to_remove,
            "/models/containers/indexes/delete",
            self.INDEX_DELETE_BATCH_SIZE,
        )

    def _remove_container_constraints(self, resource: ContainerDeploymentPlan) -> list[ChangedFieldResult]:
        return self._remove_container_fields(
            resource.constraints_to_remove,
            "/models/containers/constraints/delete",
            self.CONSTRAINT_DELETE_BATCH_SIZE,
        )

    def _remove_container_fields(
        self,
        fields_to_remove: Mapping[T_Reference, FieldChange],
        endpoint: str,
        batch_size: int,
    ) -> list[ChangedFieldResult]:
        if not fields_to_remove:
            return []
        results: list[ChangedFieldResult] = []
        for batch in chunker_sequence(list(fields_to_remove.keys()), batch_size):
            responses = self.client.http_client.request_with_retries(
                ItemsRequest(
                    endpoint_url=self.client.config.create_api_url(endpoint),
                    method="POST",
                    body=ItemIDBody(items=batch),
                )
            )
            results.extend(self._process_field_responses(responses, fields_to_remove))
        return results

    @classmethod
    def _process_resource_responses(
        cls, responses: APIResponse, change_by_id: dict[T_ResourceId, ResourceChange], endpoint: DataModelEndpoint
    ) -> list[HTTPChangeResult]:
        results: list[HTTPChangeResult] = []
        for response in responses:
            if isinstance(response, SuccessResponseItems | FailedResponseItems | FailedRequestItems):
                for id in response.ids:
                    if id not in change_by_id:
                        continue
                    results.append(HTTPChangeResult(change=change_by_id[id], http_message=response, endpoint=endpoint))
            else:
                # This should never happen as we do a ItemsRequest should always return ItemMessage responses
                raise ValueError("Bug in Neat. Got an unexpected response type.")
        return results

    @classmethod
    def _process_field_responses(
        cls, responses: APIResponse, change_by_id: Mapping[T_Reference, FieldChange]
    ) -> list[ChangedFieldResult]:
        results: list[ChangedFieldResult] = []
        for response in responses:
            if isinstance(response, SuccessResponseItems | FailedResponseItems | FailedRequestItems):
                for id in response.ids:
                    if id not in change_by_id:
                        continue
                    results.append(
                        ChangedFieldResult(resource_id=id, field_change=change_by_id[id], http_message=response)
                    )
            else:
                # This should never happen as we do a ItemsRequest should always return ItemMessage responses
                raise RuntimeError("Bug in Neat. Got an unexpected response type.")
        return results



from cognite.client import ClientConfig, CogniteClient


class NeatClientConfig(ClientConfig):
    def __init__(self, cognite_client_or_config: CogniteClient | ClientConfig):
        config = (
            cognite_client_or_config.config
            if isinstance(cognite_client_or_config, CogniteClient)
            else cognite_client_or_config
        )
        super().__init__(
            client_name=config.client_name,
            project=config.project,
            credentials=config.credentials,
            api_subversion=config.api_subversion,
            base_url=config.base_url,
            max_workers=config.max_workers,
            headers=config.headers,
            timeout=config.timeout,
            file_transfer_timeout=config.file_transfer_timeout,
            debug=config.debug,
        )

    def create_api_url(self, endpoint: str) -> str:
        """Create a full API URL for the given endpoint.

        Args:
            endpoint (str): The API endpoint to append to the base URL.

        Returns:
            str: The full API URL.

        Examples:
            >>> config = NeatClientConfig(cluster="bluefield", project="my_project", ...)
            >>> config.create_api_url("/models/instances")
            "https://bluefield.cognitedata.com/api/v1/my_project/models/instances"
        """
        if not endpoint.startswith("/"):
            endpoint = f"/{endpoint}"
        return f"{self.base_url}/api/v1/projects/{self.project}{endpoint}"



import re
import sys
from abc import ABC
from typing import Any, ClassVar, Final, Literal, TypeVar, cast, get_args

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator
from rdflib import Literal as RDFLiteral
from rdflib import URIRef

from cognite.neat._v0.core._data_model._constants import EntityTypes
from cognite.neat._v0.core._data_model.models.data_types import _XSD_TYPES, DataType
from cognite.neat._v0.core._data_model.models.entities._constants import _PARSE
from cognite.neat._v0.core._data_model.models.entities._single_value import ConceptEntity, ConceptualEntity
from cognite.neat._v0.core._issues.errors._general import NeatValueError
from cognite.neat._v0.core._utils.rdf_ import remove_namespace_from_uri

if sys.version_info <= (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


ValueConstraints = Literal["allValuesFrom", "someValuesFrom", "hasValue"]
CardinalityConstraints = Literal["minCardinality", "maxCardinality", "cardinality", "qualifiedCardinality"]


# Constants for regex patterns - more maintainable
PROPERTY_PATTERN: Final[str] = r"[a-zA-Z0-9._~?@!$&'*+,;=%-]+"
VALUE_PATTERN: Final[str] = r".+"
CARDINALITY_VALUE_PATTERN: Final[str] = r"\d+"
ON_PATTERN: Final[str] = r"[^(]*"

VALUE_CONSTRAINT_REGEX = re.compile(
    rf"^{EntityTypes.value_constraint}:(?P<property>{PROPERTY_PATTERN})\((?P<constraint>{'|'.join(get_args(ValueConstraints))}),(?P<value>{VALUE_PATTERN})\)$"
)

CARDINALITY_CONSTRAINT_REGEX = re.compile(
    rf"^{EntityTypes.cardinality_constraint}:(?P<property>{PROPERTY_PATTERN})\((?P<constraint>{'|'.join(get_args(CardinalityConstraints))}),(?P<value>{CARDINALITY_VALUE_PATTERN})(?:,(?P<on>{ON_PATTERN}))?\)$"
)


class NamedIndividualEntity(ConceptualEntity):
    type_: ClassVar[EntityTypes] = EntityTypes.named_individual

    @model_validator(mode="after")
    def reset_prefix(self) -> Self:
        self.prefix = "ni"
        return self


class ConceptPropertyRestriction(ABC, BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(
        str_strip_whitespace=True,
        arbitrary_types_allowed=True,
        strict=False,
        extra="ignore",
    )
    type_: ClassVar[EntityTypes] = EntityTypes.concept_restriction
    property_: str

    @classmethod
    def load(cls: "type[T_ConceptPropertyRestriction]", data: Any, **defaults: Any) -> "T_ConceptPropertyRestriction":
        if isinstance(data, cls):
            return data
        elif not isinstance(data, str):
            raise ValueError(f"Cannot load {cls.__name__} from {data}")

        return cls.model_validate({_PARSE: data, "defaults": defaults})

    @model_validator(mode="before")
    def _load(cls, data: Any) -> dict:
        defaults = {}
        if isinstance(data, dict) and _PARSE in data:
            defaults = data.get("defaults", {})
            data = data[_PARSE]
        if isinstance(data, dict):
            data.update(defaults)
            return data

        if not isinstance(data, str):
            raise ValueError(f"Cannot load {cls.__name__} from {data}")

        return cls._parse(data, defaults)

    @classmethod
    def _parse(cls, data: str, defaults: dict) -> dict:
        raise NotImplementedError(f"{cls.__name__} must implement _parse method")

    def dump(self) -> str:
        return self.__str__()

    def as_tuple(self) -> tuple[str, ...]:
        # We haver overwritten the serialization to str, so we need to do it manually
        extra: tuple[str, ...] = tuple(
            [
                str(v or "")
                for field_name in self.model_fields.keys()
                if (v := getattr(self, field_name)) and field_name not in {"property_"}
            ]
        )

        return self.property_, *extra

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, ConceptPropertyRestriction):
            return NotImplemented
        return self.as_tuple() < other.as_tuple()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ConceptPropertyRestriction):
            return NotImplemented
        return self.as_tuple() == other.as_tuple()

    def __hash__(self) -> int:
        return hash(str(self))


T_ConceptPropertyRestriction = TypeVar("T_ConceptPropertyRestriction", bound="ConceptPropertyRestriction")


class ConceptPropertyValueConstraint(ConceptPropertyRestriction):
    type_: ClassVar[EntityTypes] = EntityTypes.value_constraint
    constraint: ValueConstraints
    value: RDFLiteral | ConceptEntity | NamedIndividualEntity

    @field_validator("value")
    def validate_value(cls, value: Any) -> Any:
        if isinstance(value, RDFLiteral) and value.datatype is None:
            raise NeatValueError("RDFLiteral must have a datatype set, which must be one of the XSD types.")
        return value

    def __str__(self) -> str:
        value_str = (
            f"{self.value.value}^^{remove_namespace_from_uri(cast(URIRef, self.value.datatype))}"
            if isinstance(self.value, RDFLiteral)
            else str(self.value)
        )
        return f"{self.type_}:{self.property_}({self.constraint},{value_str})"

    @classmethod
    def _parse(cls, data: str, defaults: dict) -> dict:
        if not (result := VALUE_CONSTRAINT_REGEX.match(data)):
            raise NeatValueError(f"Invalid value constraint format: {data}")

        property_ = result.group("property")
        constraint = result.group("constraint")
        raw_value = result.group("value")

        value: NamedIndividualEntity | RDFLiteral | ConceptEntity
        # scenario 1: NamedIndividual as value restriction
        if raw_value.startswith("ni:"):
            value = NamedIndividualEntity.load(raw_value)
        # scenario 2: Datatype as value restriction
        elif "^^" in raw_value:
            if len(value_components := raw_value.split("^^")) == 2 and value_components[1] in _XSD_TYPES:
                value = RDFLiteral(value_components[0], datatype=DataType.load(value_components[1]).as_xml_uri_ref())
            else:
                raise NeatValueError(f"Invalid value format for datatype: {raw_value}")

        # scenario 3: ConceptEntity as value restriction
        else:
            value = ConceptEntity.load(raw_value, **defaults)

        return dict(property_=property_, constraint=constraint, value=value)


class ConceptPropertyCardinalityConstraint(ConceptPropertyRestriction):
    type_: ClassVar[EntityTypes] = EntityTypes.cardinality_constraint
    constraint: CardinalityConstraints
    value: int = Field(ge=0)
    on: DataType | ConceptEntity | None = None

    def __str__(self) -> str:
        on_str = f",{self.on}" if self.on else ""
        return f"{self.type_}:{self.property_}({self.constraint},{self.value}{on_str})"

    @classmethod
    def _parse(cls, data: str, defaults: dict) -> dict:
        if not (result := CARDINALITY_CONSTRAINT_REGEX.match(data)):
            raise NeatValueError(f"Invalid cardinality constraint format: {data}")

        property_ = result.group("property")
        constraint = result.group("constraint")
        value = result.group("value")
        on = result.group("on")
        if on:
            if on in _XSD_TYPES:
                on = DataType.load(on)
            else:
                on = cast(ConceptEntity, ConceptEntity.load(on, **defaults))

        return dict(property_=property_, constraint=constraint, value=value, on=on)


def parse_restriction(data: str, **defaults: Any) -> ConceptPropertyRestriction:
    """Parse a string to create either a value or cardinality restriction.

    Args:
        data: String representation of the restriction
        **defaults: Default values to use when parsing

    Returns:
        Either a ConceptPropertyValueConstraint or ConceptPropertyCardinalityConstraint

    Raises:
        NeatValueError: If the string cannot be parsed as either restriction type
    """
    # Check for value constraint pattern first (more specific)
    if VALUE_CONSTRAINT_REGEX.match(data):
        try:
            return ConceptPropertyValueConstraint.load(data, **defaults)
        except Exception as e:
            raise NeatValueError(f"Failed to parse value constraint: {data}") from e

    # Check for cardinality constraint pattern
    if CARDINALITY_CONSTRAINT_REGEX.match(data):
        try:
            return ConceptPropertyCardinalityConstraint.load(data, **defaults)
        except Exception as e:
            raise NeatValueError(f"Failed to parse cardinality constraint: {data}") from e

    # If neither pattern matches, provide a clear error
    raise NeatValueError(
        f"Invalid restriction format: {data}. "
        f"Expected format: '{EntityTypes.value_constraint}:property(constraint,value)' "
        f"or '{EntityTypes.cardinality_constraint}:property(constraint,value[,on])'"
    )


T_ConceptRestriction = TypeVar("T_ConceptRestriction", bound=ConceptPropertyRestriction)



"""Validators for checking containers in the data model."""

from cognite.neat._data_model.models.dms._view_property import ViewCorePropertyRequest
from cognite.neat._data_model.rules.dms._base import DataModelRule
from cognite.neat._issues import ConsistencyError

BASE_CODE = "NEAT-DMS-VIEW"


class ViewToContainerMappingNotPossible(DataModelRule):
    """Validates that container and container property referenced by view property exist.

    ## What it does
    Validates that for each view property that maps to a container and container property,
    the referenced container and container property exist.

    ## Why is this bad?
    If a view property references a container or container property that does not exist,
    the data model cannot be deployed to CDF. This means that view property will not be able to function.

    ## Example
    View WindTurbine has property location that maps to container WindTurbineContainer and property gpsCoordinates.
    If WindTurbineContainer and/or property gpsCoordinates does not exist, the data model cannot be deployed to CDF.
    """

    code = f"{BASE_CODE}-001"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        if not self.validation_resources.merged_data_model.views:
            return errors

        for view_ref in self.validation_resources.merged_data_model.views:
            view = self.validation_resources.select_view(view_ref)

            if not view:
                raise RuntimeError(
                    f"{type(self).__name__}: View {view_ref!s} not found in local resources. This is a bug in NEAT."
                )

            if view.properties is None:
                continue

            for property_ref, property_ in view.properties.items():
                if not isinstance(property_, ViewCorePropertyRequest):
                    continue

                container_ref = property_.container
                container_property = property_.container_property_identifier

                container = self.validation_resources.select_container(container_ref, container_property)

                if not container:
                    errors.append(
                        ConsistencyError(
                            message=(
                                f"View {view_ref!s} property {property_ref!s} maps to "
                                f"container {container_ref!s} which does not exist."
                            ),
                            fix="Define necessary container",
                            code=self.code,
                        )
                    )
                elif container_property not in container.properties:
                    errors.append(
                        ConsistencyError(
                            message=(
                                f"View {view_ref!s} property {property_ref!s} maps to "
                                f"container {container_ref!s} which does not have "
                                f"property '{container_property}'."
                            ),
                            fix="Define necessary container property",
                            code=self.code,
                        )
                    )

        return errors


class ImplementedViewNotExisting(DataModelRule):
    """Validates that implemented (inherited) view exists.

    ## What it does
    Validates that all views which are implemented (inherited) in the data model actually exist either locally
    or in CDF.

    ## Why is this bad?
    If a view being implemented (inherited) does not exist, the data model cannot be deployed to CDF.

    ## Example
    If view WindTurbine implements (inherits) view Asset, but Asset view does not exist in the data model
    or in CDF, the data model cannot be deployed to CDF.
    """

    code = f"{BASE_CODE}-002"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        if not self.validation_resources.merged_data_model.views:
            return errors

        for view_ref in self.validation_resources.merged_data_model.views:
            view = self.validation_resources.select_view(view_ref)

            if not view:
                raise RuntimeError(
                    f"{type(self).__name__}: View {view_ref!s} not found in local resources. This is a bug in NEAT."
                )

            if view.implements is None:
                continue
            for implement in view.implements:
                if self.validation_resources.select_view(implement) is None:
                    errors.append(
                        ConsistencyError(
                            message=f"View {view_ref!s} implements {implement!s} which is not defined.",
                            fix="Define the missing view or remove it from the implemented views list",
                            code=self.code,
                        )
                    )

        return errors


class CyclicImplements(DataModelRule):
    """Validates that view implements are not forming a cycle (i.e. cyclic graph of implements)

    ## What it does
    Runs graph analysis of the implements graph finding every cycle within the graph

    ## Why is this bad?
    You will not be able to deploy the data model to CDF, since cyclic implements are impossible to resolve
    in terms of inheritance of properties.

    ## Example
    Say we have following views: A, B, and C, where A implements B, B implements C, and C implements A. This forms
    the cyclic graph of implements A->B->C->A.

    """

    code = f"{BASE_CODE}-003"
    issue_type = ConsistencyError
    alpha = True

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        for cycle in self.validation_resources.implements_cycles:
            errors.append(
                ConsistencyError(
                    message=(
                        "Detected cycle of view implements "
                        f"{'->'.join([str(view) for view in cycle]) + f'->{cycle[0]}'}"
                    ),
                    fix="Inspect involved views and make necessary fix",
                    code=self.code,
                )
            )

        return errors



from ._base import ConceptEntity, UnknownEntity
from ._constants import Undefined, Unknown
from ._data_types import (
    AnyURI,
    Boolean,
    DataType,
    Date,
    DateTime,
    DateTimeStamp,
    Double,
    Enum,
    File,
    Float,
    Integer,
    Json,
    Long,
    Sequence,
    String,
    Timeseries,
)
from ._identifiers import URI, NameSpace
from ._parser import ParsedEntity, parse_entities, parse_entity

__all__ = [
    "URI",
    "AnyURI",
    "Boolean",
    "ConceptEntity",
    "DataType",
    "Date",
    "DateTime",
    "DateTimeStamp",
    "Double",
    "Enum",
    "File",
    "Float",
    "Integer",
    "Json",
    "Long",
    "NameSpace",
    "ParsedEntity",
    "Sequence",
    "String",
    "Timeseries",
    "Undefined",
    "Unknown",
    "UnknownEntity",
    "parse_entities",
    "parse_entity",
]



from cognite.client import ClientConfig, CogniteClient

from cognite.neat._utils.http_client import HTTPClient

from .config import NeatClientConfig
from .containers_api import ContainersAPI
from .data_model_api import DataModelsAPI
from .spaces_api import SpacesAPI
from .statistics_api import StatisticsAPI
from .views_api import ViewsAPI


class NeatClient:
    def __init__(self, cognite_client_or_config: CogniteClient | ClientConfig) -> None:
        self.config = NeatClientConfig(cognite_client_or_config)
        self.http_client = HTTPClient(self.config)
        self.data_models = DataModelsAPI(self.config, self.http_client)
        self.views = ViewsAPI(self.config, self.http_client)
        self.containers = ContainersAPI(self.config, self.http_client)
        self.spaces = SpacesAPI(self.config, self.http_client)
        self.statistics = StatisticsAPI(self.config, self.http_client)



import itertools
from typing import Any, Literal, cast, get_args

from pydantic import BaseModel, Field

from cognite.neat._data_model.deployer.data_classes import (
    AppliedChanges,
    DataModelEndpoint,
    DeploymentResult,
)


class EndpointStatistics(BaseModel):
    """Statistics for a single endpoint."""

    create: int = 0
    update: int = 0
    delete: int = 0
    skip: int = 0
    unchanged: int = 0
    failed: int = 0

    @property
    def total(self) -> int:
        """Total number of changes for this endpoint."""
        return self.create + self.update + self.delete + self.skip + self.unchanged + self.failed

    def increment(self, change_type: Literal["create", "update", "delete", "unchanged", "skip", "failed"]) -> None:
        """Increment the count for a specific change type.

        Args:
            change_type: The type of change (create, update, delete, skip, unchanged, failed).
        """
        if hasattr(self, change_type):
            setattr(self, change_type, getattr(self, change_type) + 1)
        else:
            raise RuntimeError(f"Unknown change type: {change_type}. This is a bug in NEAT.")


class ChangeTypeStatistics(BaseModel):
    """Statistics grouped by change type."""

    create: int = 0
    update: int = 0
    delete: int = 0
    skip: int = 0
    unchanged: int = 0
    failed: int = 0

    def increment(self, change_type: Literal["create", "update", "delete", "unchanged", "skip", "failed"]) -> None:
        """Increment the count for a specific change type.

        Args:
            change_type: The type of change (create, update, delete, skip, unchanged, failed).
        """
        if hasattr(self, change_type):
            setattr(self, change_type, getattr(self, change_type) + 1)
        else:
            raise RuntimeError(f"Unknown change type: {change_type}. This is a bug in NEAT.")


class SeverityStatistics(BaseModel):
    """Statistics grouped by severity."""

    SAFE: int = 0
    WARNING: int = 0
    BREAKING: int = 0

    def increment(self, severity: Literal["SAFE", "WARNING", "BREAKING"]) -> None:
        """Increment the count for a specific severity level.

        Args:
            severity: The severity level (SAFE, WARNING, BREAKING).
        """
        if hasattr(self, severity):
            setattr(self, severity, getattr(self, severity) + 1)
        else:
            raise RuntimeError(f"Unknown severity level: {severity}. This is a bug in NEAT.")


class DeploymentStatistics(BaseModel):
    """Overall deployment statistics."""

    by_endpoint: dict[str, EndpointStatistics] = Field(default_factory=dict)
    by_change_type: ChangeTypeStatistics = Field(default_factory=ChangeTypeStatistics)
    by_severity: SeverityStatistics = Field(default_factory=SeverityStatistics)

    @property
    def total_changes(self) -> int:
        """Total number of changes in the deployment."""
        return sum(endpoint_stats.total for endpoint_stats in self.by_endpoint.values())

    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        """Include computed properties in serialization."""
        data = super().model_dump(**kwargs)
        data["total_changes"] = self.total_changes
        return data

    @classmethod
    def from_deployment_result(cls, result: DeploymentResult) -> "DeploymentStatistics":
        """Create DeploymentStatistics from a DeploymentResult.

        Args:
            result: The deployment result to create statistics from.

        Returns:
            DeploymentStatistics instance with computed statistics.
        """
        stats = cls(
            by_endpoint={endpoint: EndpointStatistics() for endpoint in get_args(DataModelEndpoint)},
            by_change_type=ChangeTypeStatistics(),
            by_severity=SeverityStatistics(),
        )

        if result.is_dry_run:
            stats._update_from_dry_run(result)
        else:
            stats._update_from_deployment(result)

        return stats

    def _update_from_dry_run(self, result: DeploymentResult) -> None:
        """Update statistics from dry run mode.

        Args:
            result: The deployment result in dry run mode.
        """
        for plan in result.plan:
            for resource in plan.resources:
                self._update_single_stat(
                    endpoint=plan.endpoint,
                    change_type=resource.change_type,
                    severity=cast(Literal["SAFE", "WARNING", "BREAKING"], resource.severity.name),
                )

    def _update_from_deployment(self, result: DeploymentResult) -> None:
        """Update statistics from actual deployment mode.

        Args:
            result: The deployment result from actual deployment.
        """
        applied_changes = cast(AppliedChanges, result.responses)

        for response in itertools.chain(
            applied_changes.created,
            applied_changes.merged_updated,
            applied_changes.deletions,
            applied_changes.unchanged,
            applied_changes.skipped,
        ):
            self._update_single_stat(
                endpoint=response.endpoint,
                change_type=response.change.change_type if response.is_success else "failed",
                severity=cast(Literal["SAFE", "WARNING", "BREAKING"], response.change.severity.name),
            )

    def _update_single_stat(
        self,
        endpoint: str,
        change_type: Literal["create", "update", "delete", "unchanged", "skip", "failed"],
        severity: Literal["SAFE", "WARNING", "BREAKING"],
    ) -> None:
        """Update all statistics for a single change.

        Args:
            endpoint: The endpoint type (spaces, containers, views, datamodels).
            change_type: The type of change (create, update, delete, skip, unchanged, failed).
            severity: The severity level (SAFE, WARNING, BREAKING).
        """
        # Update by change type statistics
        self.by_change_type.increment(change_type)

        # Update by endpoint statistics
        if endpoint in self.by_endpoint:
            self.by_endpoint[endpoint].increment(change_type)
        else:
            raise RuntimeError(f"Unknown endpoint: {endpoint}. This is a bug in NEAT.")

        # Update severity statistics
        self.by_severity.increment(severity)



import itertools
import json
from collections import defaultdict
from collections.abc import Iterable
from pathlib import Path
from typing import cast

import yaml
from cognite.client import data_modeling as dm
from cognite.client.data_classes.capabilities import Capability, DataModelsAcl
from rdflib import URIRef

from cognite.neat._v0.core._client import NeatClient
from cognite.neat._v0.core._client._api.data_modeling_loaders import MultiCogniteAPIError
from cognite.neat._v0.core._constants import COGNITE_SPACES
from cognite.neat._v0.core._issues import IssueList, NeatIssue
from cognite.neat._v0.core._issues.errors import ResourceCreationError, ResourceNotFoundError
from cognite.neat._v0.core._issues.warnings import NeatValueWarning
from cognite.neat._v0.core._store import NeatInstanceStore
from cognite.neat._v0.core._utils.collection_ import iterate_progress_bar_if_above_config_threshold
from cognite.neat._v0.core._utils.rdf_ import namespace_as_space, split_uri
from cognite.neat._v0.core._utils.text import NamingStandardization
from cognite.neat._v0.core._utils.upload import UploadResult

from ._base import _END_OF_CLASS, _START_OF_CLASS, CDFLoader


class InstanceSpaceLoader(CDFLoader[dm.SpaceApply]):
    """Loads Instance Space into Cognite Data Fusion (CDF).

    There are three ways to determine the space for each instance:
    1. If `instance_space` is provided, all instances will be assigned to that space, i.e., the space
        is constant for all instances.
    (If not it is set based on the triples (subject, predicate, object) in the graph store.)
    2. If `space_property` is provided, this is the predicate and the space is set to the object. The `instance_space`
       is used as a fallback if the object is not a valid space.
    3. If `use_source_space` is set to True, the instances are assumed to be extracted from CDF and the space is part
        of the subject.

    This class exposes the `space_by_instance_uri` property used by the DMSLoader to lookup space for each instance URI.

    Args:
        graph_store (NeatInstanceStore): The graph store to load the data from.
        instance_space (str): The instance space to load the data into.
        space_property (str): The property to use to determine the space for each instance.
        use_source_space (bool): If True, use the source space of the instances when extracted from CDF.
        neat_prefix_by_predicate_uri (dict[URIRef, str] | None): A dictionary that maps a predicate URIRef to a
            prefix that Neat added to the object upon extraction. This is used to remove the prefix from the
            object before creating the instance.
    """

    def __init__(
        self,
        graph_store: NeatInstanceStore | None = None,
        instance_space: str | None = None,
        space_property: str | None = None,
        use_source_space: bool = False,
        neat_prefix_by_predicate_uri: dict[URIRef, str] | None = None,
    ) -> None:
        self.graph_store = graph_store
        self.instance_space = instance_space
        self.space_property = space_property
        self.use_source_space = use_source_space
        self.neat_prefix_by_predicate_uri = neat_prefix_by_predicate_uri or {}

        self._lookup_issues = IssueList()

        self._has_looked_up = False
        # This is a dictionary mapping instance URIs to their respective spaces
        # This is exposed through the property space_by_instance_uri. If the instance_space or space_property is
        # set (1. and 2.) this is changed to a defaultdict with the instance_space as the default value.
        self._space_by_instance_uri: dict[URIRef, str] = {}

    @property
    def space_by_instance_uri(self) -> dict[URIRef, str]:
        """Returns a dictionary mapping instance URIs to their respective spaces."""
        if not self._has_looked_up:
            self._lookup_spaces()
            self._has_looked_up = True
        return self._space_by_instance_uri

    def _get_required_capabilities(self) -> list[Capability]:
        return [
            DataModelsAcl(
                actions=[
                    DataModelsAcl.Action.Write,
                    DataModelsAcl.Action.Read,
                ],
                scope=DataModelsAcl.Scope.All(),
            )
        ]

    def _upload_to_cdf(
        self,
        client: NeatClient,
        spaces: list[dm.SpaceApply],
        dry_run: bool,
        read_issues: IssueList,
        class_name: str | None = None,
    ) -> Iterable[UploadResult]:
        cdf_spaces = client.data_modeling.spaces.retrieve([space.space for space in spaces])
        cdf_space_by_id = {item.space: item for item in cdf_spaces}

        to_create = dm.SpaceApplyList([])
        to_update = dm.SpaceApplyList([])
        unchanged = dm.SpaceApplyList([])

        for local_space in spaces:
            cdf_space = cdf_space_by_id.get(local_space.space)
            if cdf_space is None:
                to_create.append(local_space)
            elif cdf_space != local_space.as_write():
                to_update.append(local_space)
            else:
                unchanged.append(local_space)
        loader = client.loaders.spaces
        results: UploadResult[str] = UploadResult("instance spaces")
        results.unchanged.update(unchanged.as_ids())
        if dry_run:
            results.created.update(to_create.as_ids())
            results.changed.update(to_update.as_ids())
            yield results
        if to_create:
            try:
                client.loaders.spaces.create(to_create)
            except MultiCogniteAPIError as e:
                results.failed_created.update(to_create.as_ids())
                for error in e.errors:
                    results.error_messages.append(f"Failed to create {loader.resource_name}: {error!s}")
            else:
                results.created.update(to_create.as_ids())

        if to_update:
            try:
                client.loaders.spaces.update(to_update)
            except MultiCogniteAPIError as e:
                results.failed_changed.update(to_update.as_ids())
                for error in e.errors:
                    results.error_messages.append(f"Failed to update {loader.resource_name}: {error!s}")
            else:
                results.changed.update(to_update.as_ids())

        yield results

    def write_to_file(self, filepath: Path) -> None:
        """Dumps the instance spaces to file."""
        if filepath.suffix not in [".json", ".yaml", ".yml"]:
            raise ValueError(f"File format {filepath.suffix} is not supported")
        dumped: dict[str, list] = {"spaces": [], "issues": []}
        for item in self.load(stop_on_exception=False):
            key = {
                dm.SpaceApply: "spaces",
                NeatIssue: "issues",
            }.get(type(item))
            if key is None:
                # This should never happen, and is a bug in neat
                raise ValueError(f"Item {item} is not supported. This is a bug in neat please report it.")
            dumped[key].append(item.dump())
        with filepath.open("w", encoding=self._encoding, newline=self._new_line) as f:
            if filepath.suffix == ".json":
                json.dump(dumped, f, indent=2)
            else:
                yaml.safe_dump(dumped, f, sort_keys=False)

    def _load(
        self, stop_on_exception: bool = False
    ) -> Iterable[dm.SpaceApply | NeatIssue | type[_END_OF_CLASS] | _START_OF_CLASS]:
        self._lookup_spaces()
        if self._lookup_issues.has_errors and stop_on_exception:
            raise self._lookup_issues.as_errors()
        yield from self._lookup_issues
        seen: set[str] = set()
        for space_str in set(self.space_by_instance_uri.values()):
            if space_str in seen or space_str in COGNITE_SPACES:
                continue
            yield dm.SpaceApply(space=space_str)
            seen.add(space_str)

    def _lookup_spaces(self) -> None:
        # Case 1: Same instance space for all instances:
        if isinstance(self.instance_space, str) and self.space_property is None and self.use_source_space is False:
            self._space_by_instance_uri = defaultdict(lambda: cast(str, self.instance_space))
            # Adding a dummy entry to ensure that the instance space is included
            self._space_by_instance_uri[URIRef(self.instance_space)] = self.instance_space
            return
        if self.graph_store is None:
            raise ValueError("Graph store must be provided to lookup spaces")
        # Case 3: Use the source space, i.e., the space of the instances when extracted from CDF
        if self.use_source_space and self.instance_space is None and self.space_property is None:
            self._lookup_space_via_instance_uris(self.graph_store)
        # Case 2: Use a property on each instance to determine the space.
        elif self.space_property is not None and self.use_source_space is False:
            if self.instance_space is None:
                raise ValueError(
                    "Missing fallback instance space. This is required when "
                    f"using space_property='{self.space_property}'"
                )
            self._space_by_instance_uri = defaultdict(lambda: cast(str, self.instance_space))
            self._lookup_space_via_property(self.graph_store, self.space_property)
        else:
            raise ValueError("Either 'instance_space', 'space_property', or 'use_source_space' must be provided.")

    def _lookup_space_via_instance_uris(self, graph_store: NeatInstanceStore) -> None:
        instance_iterable = itertools.chain(
            (res[0] for res in graph_store.queries.select.list_instances_ids()),
            graph_store.queries.select.list_instance_object_ids(),
        )

        for instance_uri in instance_iterable:
            namespace, external_id = split_uri(instance_uri)
            space = namespace_as_space(namespace)
            if space is None:
                error = ResourceCreationError(instance_uri, "instance", "This instance was not extracted from CDF.")
                self._lookup_issues.append(error)
            else:
                self._space_by_instance_uri[instance_uri] = space

    def _lookup_space_via_property(self, graph_store: NeatInstanceStore, space_property: str) -> None:
        properties_by_uriref = graph_store.queries.select.properties()
        space_property_uri = next((k for k, v in properties_by_uriref.items() if v == space_property), None)
        if space_property_uri is None:
            error: ResourceNotFoundError[str, str] = ResourceNotFoundError(
                self.space_property,
                "property",
                more=f"Could not find the {space_property} in the graph.",
            )
            self._lookup_issues.append(error)
            return

        class_with_total_pair = graph_store.queries.select.summarize_instances()
        total = sum([count for _, count in class_with_total_pair])
        instance_iterable = graph_store.queries.select.list_instances_ids_by_space(space_property_uri)
        instance_iterable = iterate_progress_bar_if_above_config_threshold(
            instance_iterable, total, f"Looking up spaces for {total} instances..."
        )
        neat_prefix = self.neat_prefix_by_predicate_uri.get(space_property_uri)
        warned_spaces: set[str] = set()
        for instance, space in instance_iterable:
            if neat_prefix:
                space = space.removeprefix(neat_prefix)

            clean_space = NamingStandardization.standardize_space_str(space)
            if clean_space != space and space not in warned_spaces:
                self._lookup_issues.append(
                    NeatValueWarning(f"Invalid space in property {space_property}: {space}. Fixed to {clean_space}")
                )
                warned_spaces.add(space)

            self._space_by_instance_uri[instance] = clean_space



from collections.abc import Callable, Mapping
from dataclasses import dataclass, field
from typing import Literal

from pydantic_core import ErrorDetails


def as_json_path(loc: tuple[str | int, ...]) -> str:
    """Converts a location tuple to a JSON path.

    Args:
        loc: The location tuple to convert.

    Returns:
        A JSON path string.
    """
    if not loc:
        return ""
    # +1 to convert from 0-based to 1-based indexing
    prefix = ""
    if isinstance(loc[0], int):
        prefix = "item"

    suffix = ".".join([str(x) if isinstance(x, str) else f"[{x + 1}]" for x in loc]).replace(".[", "[")
    return f"{prefix}{suffix}"


@dataclass
class ValidationContext:
    """
    Context for validation errors providing configuration for error message formatting.

    This class configures how validation errors are reported, including location formatting,
    field naming conventions, and how to present missing required fields.

    Attributes:
        parent_loc: Optional location tuple to prepend to each error location.
            This is useful when the error is for a nested model and you want to include the location
            of the parent model.
        humanize_location: A function that converts a location tuple to a human-readable string.
            The default is `as_json_path`, which converts the location to a JSON path.
            This can for example be replaced when the location comes from an Excel table.
        field_name: The name use for "field" in error messages. Default is "field". This can be changed to
            "column" or "value" to better fit the context.
        field_renaming: Optional mapping of field names to source names.
            This is useful when the field names in the model are different from the names in the source.
            For example, if the model field is "asset_id" but the source column is "Asset ID",
            you can provide a mapping {"asset_id": "Asset ID"} to have the error messages use the source names.
        missing_required_descriptor: How to describe missing required fields. Default is "missing".
            Other option is "empty" which can be more suitable for table data.
    """

    parent_loc: tuple[int | str, ...] = field(default_factory=tuple)
    humanize_location: Callable[[tuple[int | str, ...]], str] = as_json_path
    field_name: Literal["field", "column", "value"] = "field"
    field_renaming: Mapping[str, str] = field(default_factory=dict)
    missing_required_descriptor: Literal["empty", "missing"] = "missing"


def humanize_validation_error(
    error: ErrorDetails,
    context: ValidationContext | None = None,
) -> str:
    """Converts a pydantic ErrorDetails object to a human-readable format.
    This overwrites the default error messages from Pydantic to be better suited for NEAT users.
    Args:
        error: The ErrorDetails object to convert.
        context: The context for humanizing the error.
    Returns:
        A human-readable error message.
    """

    context = context or ValidationContext()

    loc = (*context.parent_loc, *error["loc"])
    type_ = error["type"]

    if type_ == "missing":
        msg = f"Missing required {context.field_name}: {loc[-1]!r}"
    elif type_ == "extra_forbidden":
        msg = f"Unused {context.field_name}: {loc[-1]!r}"
    elif type_ == "value_error":
        msg = str(error["ctx"]["error"])
    elif type_ == "literal_error":
        msg = f"{error['msg']}. Got {error['input']!r}."
    elif type_ == "string_type":
        msg = f"{error['msg']}. Got {error['input']!r} of type {type(error['input']).__name__}. "
    elif type_ == "model_type":
        model_name = error["ctx"].get("class_name", "unknown")
        msg = (
            f"Input must be an object of type {model_name}. Got {error['input']!r} of "
            f"type {type(error['input']).__name__}."
        )
    elif type_ == "union_tag_invalid":
        msg = error["msg"].replace(", 'direct'", "").replace("found using 'type' ", "").replace("tag", "value")
    elif type_ == "string_pattern_mismatch":
        msg = f"string '{error['input']}' does not match the required pattern: '{error['ctx']['pattern']}'."

    elif type_.endswith("_type"):
        msg = f"{error['msg']}. Got {error['input']!r} of type {type(error['input']).__name__}."
    else:
        # Default to the Pydantic error message
        msg = error["msg"]

    if type_.endswith("dict_type") and len(loc) > 1:
        # If this is a dict_type error for a JSON field, the location will be:
        #  dict[str,json-or-python[json=any,python=tagged-union[list[...],dict[str,...],str,bool,int,float,none]]]
        #  This is hard to read, so we simplify it to just the field name.
        loc = tuple(["dict" if isinstance(x, str) and "json-or-python" in x else x for x in loc])

    error_suffix = f"{msg[:1].casefold()}{msg[1:]}"

    if len(loc) >= 3 and context.field_name == "column" and loc[-3:] == ("type", "enum", "values"):
        # Special handling for enum errors in table columns
        msg = _enum_message(type_, loc, context)
    elif len(loc) > 1 and type_ in {"extra_forbidden", "missing"}:
        if context.missing_required_descriptor == "empty" and type_ == "missing":
            # This is a table so we modify the error message.
            msg = (
                f"In {context.humanize_location(loc[:-1])} the {context.field_name}"
                f" {context.field_renaming.get(str(loc[-1]), loc[-1])!r} "
                "cannot be empty."
            )
        else:
            # We skip the last element as this is in the message already
            msg = f"In {context.humanize_location(loc[:-1])} {error_suffix.replace('field', context.field_name)}"
    elif len(loc) > 1:
        if context.parent_loc == ("Metadata",) and len(loc) == 2:
            msg = f"In table '{loc[0]}' '{loc[1]}' {error_suffix}"
        else:
            msg = f"In {context.humanize_location(loc)} {error_suffix}"
    elif len(loc) == 1 and isinstance(loc[0], str) and type_ not in {"extra_forbidden", "missing"}:
        msg = f"In {context.field_name} {loc[0]!r}, {error_suffix}"

    msg = msg.strip()
    if not msg.endswith("."):
        msg += "."
    return msg


def _enum_message(type_: str, loc: tuple[int | str, ...], context: ValidationContext) -> str:
    """Special handling of enum errors in table columns."""

    if loc[-1] != "values":
        raise RuntimeError("This is a neat bug, report to the team.")
    if type_ == "missing":
        return (
            f"In {context.humanize_location(loc[:-1])} definition should include "
            "a reference to a collection in the 'Enum' sheet (e.g., collection='MyEnumCollection')."
        )
    elif type_ == "too_short":
        return f"In {context.humanize_location(loc[:-1])} collection is not defined in the 'Enum' sheet"
    else:
        raise RuntimeError("This is a neat bug, report to the team.")



from pydantic import BaseModel


class _UndefinedType(BaseModel): ...


class _UnknownType(BaseModel):
    def __str__(self) -> str:
        return "#N/A"

    def __hash__(self) -> int:
        return hash(str(self))


# This is a trick to make Undefined and Unknown singletons
Undefined = _UndefinedType()
Unknown = _UnknownType()


PREFIX_PATTERN = r"^[a-zA-Z][a-zA-Z0-9_-]{0,41}[a-zA-Z0-9]?$"
SUFFIX_PATTERN = r"^[a-zA-Z0-9._~?@!$&'*+,;=%-]+$"
VERSION_PATTERN = r"^[a-zA-Z0-9]([.a-zA-Z0-9_-]{0,41}[a-zA-Z0-9])?$"



from cognite.neat._v0.core._client.data_classes.schema import DMSSchema

from ._unverified import (
    UnverifiedPhysicalContainer,
    UnverifiedPhysicalDataModel,
    UnverifiedPhysicalEnum,
    UnverifiedPhysicalMetadata,
    UnverifiedPhysicalNodeType,
    UnverifiedPhysicalProperty,
    UnverifiedPhysicalView,
)
from ._validation import PhysicalValidation
from ._verified import (
    PhysicalContainer,
    PhysicalDataModel,
    PhysicalEnum,
    PhysicalMetadata,
    PhysicalNodeType,
    PhysicalProperty,
    PhysicalView,
)

__all__ = [
    "DMSSchema",
    "PhysicalContainer",
    "PhysicalDataModel",
    "PhysicalEnum",
    "PhysicalMetadata",
    "PhysicalNodeType",
    "PhysicalProperty",
    "PhysicalValidation",
    "PhysicalView",
    "UnverifiedPhysicalContainer",
    "UnverifiedPhysicalDataModel",
    "UnverifiedPhysicalEnum",
    "UnverifiedPhysicalMetadata",
    "UnverifiedPhysicalNodeType",
    "UnverifiedPhysicalProperty",
    "UnverifiedPhysicalView",
]



from cognite.neat._client.init.main import get_cognite_client

from ._config import NeatConfig, get_neat_config_from_file
from ._session import NeatSession
from ._version import __version__

__all__ = ["NeatConfig", "NeatSession", "__version__", "get_cognite_client", "get_neat_config_from_file"]



from abc import ABC, abstractmethod

from cognite.neat._data_model.models.dms import RequestSchema


class DMSImporter(ABC):
    """This is the base class for all DMS importers."""

    @abstractmethod
    def to_data_model(self) -> RequestSchema:
        """Convert the imported data to a RequestSchema.

        Returns:
            RequestSchema: The data model as a RequestSchema.
        """
        raise NotImplementedError()



from typing import Literal, overload

from cognite.neat._v0.core._data_model.models.data_types import DataType
from cognite.neat._v0.core._issues.errors import NeatTypeError

from ._multi_value import MultiValueTypeInfo
from ._single_value import (
    ConceptEntity,
    EdgeEntity,
    PhysicalUnknownEntity,
    ReverseConnectionEntity,
    Unknown,
    UnknownEntity,
    ViewEntity,
)


@overload
def load_value_type(
    raw: str | MultiValueTypeInfo | DataType | ConceptEntity | UnknownEntity,
    default_prefix: str,
    return_on_failure: Literal[False] = False,
) -> MultiValueTypeInfo | DataType | ConceptEntity | UnknownEntity: ...


@overload
def load_value_type(
    raw: str | MultiValueTypeInfo | DataType | ConceptEntity | UnknownEntity,
    default_prefix: str,
    return_on_failure: Literal[True],
) -> MultiValueTypeInfo | DataType | ConceptEntity | UnknownEntity | None | str: ...


def load_value_type(
    raw: str | MultiValueTypeInfo | DataType | ConceptEntity | UnknownEntity,
    default_prefix: str,
    return_on_failure: Literal[True, False] = False,
) -> MultiValueTypeInfo | DataType | ConceptEntity | UnknownEntity | None | str:
    """
    Loads a value type from a raw string or entity.

    Args:
        raw: The raw value to load.
        default_prefix: The default prefix to use if not specified in the raw value.
        return_on_failure: If True, returns the raw value on parsing failure instead of raising an error.

    Returns:
        The loaded value type entity, or the raw value if loading fails and `return_on_failure` is True.
    """
    if isinstance(raw, MultiValueTypeInfo | DataType | ConceptEntity | UnknownEntity):
        return raw
    elif isinstance(raw, str):
        # property holding xsd data type
        # check if it is multi value type
        if "," in raw:
            value_type = MultiValueTypeInfo.load(raw)
            value_type.set_default_prefix(default_prefix)
            return value_type
        elif DataType.is_data_type(raw):
            return DataType.load(raw)

        # unknown value type
        elif raw == str(Unknown):
            return UnknownEntity()

        # property holding link to class
        else:
            return ConceptEntity.load(raw, prefix=default_prefix, return_on_failure=return_on_failure)
    else:
        raise NeatTypeError(f"Invalid value type: {type(raw)}")


@overload
def load_dms_value_type(
    raw: str | DataType | ViewEntity | PhysicalUnknownEntity,
    default_space: str,
    default_version: str,
    return_on_failure: Literal[False],
) -> DataType | ViewEntity | PhysicalUnknownEntity: ...


@overload
def load_dms_value_type(
    raw: str | DataType | ViewEntity | PhysicalUnknownEntity,
    default_space: str,
    default_version: str,
    return_on_failure: Literal[True],
) -> DataType | ViewEntity | PhysicalUnknownEntity | str: ...


def load_dms_value_type(
    raw: str | DataType | ViewEntity | PhysicalUnknownEntity,
    default_space: str,
    default_version: str,
    return_on_failure: Literal[True, False] = False,
) -> DataType | ViewEntity | PhysicalUnknownEntity | str:
    """
    Loads a value type from a raw string or entity in the context of a data modeling service

    Args:
        raw: The raw value to load.
        default_space: The default space to use if not specified in the raw value.
        default_version: The default version to use if not specified in the raw value.
        return_on_failure: If True, returns the raw value on parsing failure instead of raising an error.

    Returns:
        The loaded value type entity, or the raw value if loading fails and `return_on_failure` is True.
    """
    if isinstance(raw, DataType | ViewEntity | PhysicalUnknownEntity):
        return raw
    elif isinstance(raw, str):
        if DataType.is_data_type(raw):
            return DataType.load(raw)
        elif raw == str(Unknown):
            return PhysicalUnknownEntity()
        else:
            return ViewEntity.load(
                raw, space=default_space, version=default_version, return_on_failure=return_on_failure
            )
    raise NeatTypeError(f"Invalid value type: {type(raw)}")


@overload
def load_connection(
    raw: Literal["direct"] | ReverseConnectionEntity | EdgeEntity | str | None,
    default_space: str,
    default_version: str,
    return_on_failure: Literal[False] = False,
) -> Literal["direct"] | ReverseConnectionEntity | EdgeEntity | None: ...


@overload
def load_connection(
    raw: Literal["direct"] | ReverseConnectionEntity | EdgeEntity | str | None,
    default_space: str,
    default_version: str,
    return_on_failure: Literal[True],
) -> Literal["direct"] | ReverseConnectionEntity | EdgeEntity | None | str: ...


def load_connection(
    raw: Literal["direct"] | ReverseConnectionEntity | EdgeEntity | str | None,
    default_space: str,
    default_version: str,
    return_on_failure: Literal[True, False] = False,
) -> Literal["direct"] | ReverseConnectionEntity | EdgeEntity | None | str:
    if isinstance(raw, str) and raw.lower() == "direct":
        return "direct"  # type: ignore[return-value]
    elif isinstance(raw, EdgeEntity | ReverseConnectionEntity) or raw is None:
        return raw  # type: ignore[return-value]
    elif isinstance(raw, str) and raw.startswith("edge"):
        return EdgeEntity.load(raw, space=default_space, version=default_version, return_on_failure=return_on_failure)  # type: ignore[return-value]
    elif isinstance(raw, str) and raw.startswith("reverse"):
        return ReverseConnectionEntity.load(raw, return_on_failure=return_on_failure)  # type: ignore[return-value]
    raise NeatTypeError(f"Invalid connection: {type(raw)}")



from ._unverified import (
    UnverifiedConcept,
    UnverifiedConceptualDataModel,
    UnverifiedConceptualMetadata,
    UnverifiedConceptualProperty,
)
from ._validation import ConceptualValidation
from ._verified import (
    Concept,
    ConceptualDataModel,
    ConceptualMetadata,
    ConceptualProperty,
)

__all__ = [
    "Concept",
    "ConceptualDataModel",
    "ConceptualMetadata",
    "ConceptualProperty",
    "ConceptualValidation",
    "UnverifiedConcept",
    "UnverifiedConceptualDataModel",
    "UnverifiedConceptualMetadata",
    "UnverifiedConceptualProperty",
]



from collections.abc import Callable
from functools import wraps
from typing import Any, Protocol, TypeVar

from cognite.neat._issues import ConsistencyError, Recommendation
from cognite.neat._session._usage_analytics._collector import Collector
from cognite.neat._store._store import NeatStore
from cognite.neat._utils.text import NEWLINE, split_on_capitals


class HasStore(Protocol):
    _store: NeatStore


T_Class = TypeVar("T_Class", bound=HasStore)

_COLLECTOR = Collector()


def session_wrapper(cls: type[T_Class]) -> type[T_Class]:
    # 1. Define the method decorator inside
    def _handle_method_call(func: Callable[..., Any]) -> Callable[..., Any]:
        """Decorator to handle exceptions and print provenance length"""

        @wraps(func)
        def wrapper(self: HasStore, *args: Any, **kwargs: Any) -> Any:
            display_name = f"{' '.join(split_on_capitals(cls.__name__))} - {func.__name__}"
            identifier = f"{cls.__name__}.{func.__name__}"
            try:
                res = func(self, *args, **kwargs)
                if not self._store.provenance or ("DataModel" not in identifier and "CDF" not in identifier):
                    print(f"{display_name} ✅")
                    if _COLLECTOR.can_collect:
                        _COLLECTOR.collect("action", {"action": identifier, "success": True})
                    return res
                change = self._store.provenance[-1]

                recommendation_count = (
                    len(recommendations)
                    if change.issues and (recommendations := change.issues.by_type().get(Recommendation))
                    else 0
                )
                consistency_errors_count = (
                    len(consistency_errors)
                    if change.issues and (consistency_errors := change.issues.by_type().get(ConsistencyError))
                    else 0
                )
                syntax_errors_count = len(change.errors) if change.errors else 0
                errors_count = consistency_errors_count + syntax_errors_count
                total_insights = recommendation_count + consistency_errors_count + syntax_errors_count

                data_model_not_read = not change.successful and "ReadPhysicalDataModel" in identifier

                if not change.successful:
                    success_icon = "❌"
                elif change.successful and consistency_errors_count:
                    success_icon = "⚠️"
                else:
                    success_icon = "✅"

                print(
                    f"{display_name} "
                    f"{success_icon}"
                    f"{f' | Insights: {total_insights} (of which {errors_count} errors)' if total_insights > 0 else ''}"
                    f"{NEWLINE + '⚠️ Data model not read into session' if data_model_not_read else ''}"
                    f"{NEWLINE + '📋 For details on issues run .issues' if change.issues or change.errors else ''}"
                    f"{NEWLINE + '📊 For details on result run .result' if change.result else ''}"
                )
                if _COLLECTOR.can_collect:
                    event = change.as_mixpanel_event()
                    event["action"] = identifier
                    _COLLECTOR.collect("action", event)
                    if change.result:
                        event = change.result.as_mixpanel_event()
                        event["action"] = identifier
                        _COLLECTOR.collect("deployment", event)

                return res

            # if an error occurs, we catch it and print it out instead of
            # getting a full traceback
            except Exception as e:
                print(f"{display_name} ❌")
                print(f"{e!s}")
                if _COLLECTOR.can_collect:
                    _COLLECTOR.collect("action", {"action": identifier, "success": False, "error_message": str(e)})

        return wrapper

    # Iterate through all attributes of the class
    for attr_name in dir(cls):
        # Skip private/protected methods (starting with _)
        if not attr_name.startswith("_"):
            attr = getattr(cls, attr_name)
            # Only wrap callable methods
            if callable(attr):
                # Replace the original method with wrapped version
                setattr(cls, attr_name, _handle_method_call(attr))
    # Intercept __init__ to wrap any methods added via setattr
    original_init = cls.__init__

    @wraps(original_init)
    def pick_alpha_methods(self: HasStore, *args: Any, **kwargs: Any) -> Any:
        """This method wraps any instance methods added during __init__. which is the case for alpha methods"""
        original_init(self, *args, **kwargs)
        # Wrap any instance methods added during init
        for attr_name in dir(self):
            if not attr_name.startswith("_"):
                attr = getattr(self, attr_name, None)
                # Check if it's an instance method (not from the class)
                if callable(attr) and attr_name not in vars(self.__class__):
                    # Wrap and set on the instance
                    wrapped = (
                        _handle_method_call(attr.__func__) if hasattr(attr, "__func__") else _handle_method_call(attr)
                    )
                    setattr(self, attr_name, wrapped.__get__(self, type(self)))

    cls.__init__ = pick_alpha_methods  # type: ignore[assignment]
    return cls



from pydantic import BaseModel, ConfigDict, Field

from cognite.neat._data_model.deployer.data_classes import FieldChange
from cognite.neat._data_model.models.dms import SchemaResourceId


class FixAction(BaseModel):
    """An atomic, individually-applicable fix for a schema issue.

    Attributes:
        resource_id: Reference to the resource being modified.
        changes: List of field-level changes.
        message: Human-readable description of what this fix does.
        code: The validator code (e.g., "NEAT-DMS-PERFORMANCE-001") for grouping in UI.
    """

    model_config = ConfigDict(frozen=True)

    resource_id: SchemaResourceId
    changes: tuple[FieldChange, ...] = Field(default_factory=tuple)
    message: str | None = None
    code: str



from collections.abc import Iterable, Set
from typing import Literal

from cognite.neat._v0.core._data_model.models import ConceptualDataModel, SheetList
from cognite.neat._v0.core._data_model.models.conceptual import Concept, ConceptualProperty
from cognite.neat._v0.core._data_model.models.data_types import DataType
from cognite.neat._v0.core._data_model.models.entities import (
    ConceptEntity,
    MultiValueTypeInfo,
    UnknownEntity,
)
from cognite.neat._v0.core._data_model.transformers import VerifiedDataModelTransformer


class UnionConceptualDataModel(VerifiedDataModelTransformer[ConceptualDataModel, ConceptualDataModel]):
    """Takes the union two conceptual models.
    Args:
        primary: The primary model to merge with the secondary model given in the transform method.
    """

    def __init__(self, primary: ConceptualDataModel) -> None:
        self.primary = primary

    def transform(self, data_model: ConceptualDataModel) -> ConceptualDataModel:
        primary_model = self.primary
        secondary_model = data_model

        output = primary_model.model_copy(deep=True)
        secondary_concepts = {cls.concept: cls for cls in secondary_model.concepts}
        secondary_properties = {(prop.concept, prop.property_): prop for prop in secondary_model.properties}

        union_concepts_by_id = self._union_concepts(output.concepts, secondary_concepts)
        output.concepts = SheetList[Concept](union_concepts_by_id.values())

        union_properties = self._union_properties(
            output.properties, secondary_properties, set(union_concepts_by_id.keys())
        )
        output.properties = SheetList[ConceptualProperty](union_properties.values())

        return output

    def _union_concepts(
        self, primary_concepts: Iterable[Concept], new_concepts: dict[ConceptEntity, Concept]
    ) -> dict[ConceptEntity, Concept]:
        union_concepts = {cls.concept: cls for cls in primary_concepts}
        for concept, primary_concept in union_concepts.items():
            if concept not in new_concepts:
                continue
            secondary_concept = new_concepts[concept]
            union_concepts[concept] = self.union_concepts(
                primary=primary_concept,
                secondary=secondary_concept,
                conflict_resolution="combined",
            )

        for concept, secondary_concept in new_concepts.items():
            if concept not in union_concepts:
                union_concepts[concept] = secondary_concept
        return union_concepts

    def _union_properties(
        self,
        primary_properties: Iterable[ConceptualProperty],
        secondary_properties: dict[tuple[ConceptEntity, str], ConceptualProperty],
        used_concepts: Set[ConceptEntity],
    ) -> dict[tuple[ConceptEntity, str], ConceptualProperty]:
        union_properties = {(prop.concept, prop.property_): prop for prop in primary_properties}
        for (concept, prop_id), primary_property in union_properties.items():
            if (concept not in used_concepts) or (concept, prop_id) not in secondary_properties:
                continue
            secondary_property = secondary_properties[(concept, prop_id)]
            union_properties[(concept, prop_id)] = self.union_properties(
                primary=primary_property,
                secondary=secondary_property,
                conflict_resolution="combined",
            )

        for (concept, prop_id), prop in secondary_properties.items():
            if (concept, prop_id) not in union_properties and concept in used_concepts:
                union_properties[(concept, prop_id)] = prop
        return union_properties

    @classmethod
    def union_concepts(
        cls,
        primary: Concept,
        secondary: Concept,
        conflict_resolution: Literal["priority", "combined"] = "priority",
    ) -> Concept:
        """Union two concepts.

        Args:
            primary (Concept): The primary concept.
            secondary (Concept): The secondary concept.
            conflict_resolution (Literal["priority", "combined"]): How to resolve conflicts:
                - "priority": Keep the primary concept with fallback to the secondary.
                - "combined": Merge implements from both concepts. (only applies to implements)
        Returns:
            Concept: The union of the two concepts.

        """
        if conflict_resolution == "combined":
            all_implements = (primary.implements or []) + (secondary.implements or [])
            implements = list(dict.fromkeys(all_implements))
        else:
            implements = (primary.implements or secondary.implements or []).copy()

        return Concept(
            neatId=primary.neatId,
            concept=primary.concept,
            name=primary.name or secondary.name,
            description=primary.description or secondary.description,
            implements=implements,
            instance_source=primary.instance_source or secondary.instance_source,
            physical=primary.physical,
        )

    @classmethod
    def union_properties(
        cls,
        primary: ConceptualProperty,
        secondary: ConceptualProperty,
        conflict_resolution: Literal["priority", "combined"] = "priority",
    ) -> ConceptualProperty:
        """Union two conceptual properties.

        Args:
            primary (ConceptualProperty): The primary property.
            secondary (ConceptualProperty): The secondary property.
            conflict_resolution (Literal["priority", "combined"]): How to resolve conflicts:
                - "priority": Keep the primary property with fallback to the secondary.
                - "combined": Merge value types and instance sources.
        Returns:
            ConceptualProperty: The union of the two properties.

        """
        if conflict_resolution == "combined":
            all_sources = (primary.instance_source or []) + (secondary.instance_source or [])
            instance_source = list(dict.fromkeys(all_sources))
        else:
            instance_source = (primary.instance_source or secondary.instance_source or []).copy()

        use_primary = conflict_resolution == "priority"
        return ConceptualProperty(
            neatId=primary.neatId,
            concept=primary.concept,
            property_=primary.property_,
            name=primary.name or secondary.name,
            description=primary.description or secondary.description,
            min_count=primary.min_count
            if use_primary
            else cls._union_min_count(primary.min_count, secondary.min_count),
            max_count=primary.max_count
            if use_primary
            else cls._union_max_count(primary.max_count, secondary.max_count),
            default=primary.default or secondary.default,
            value_type=primary.value_type
            if use_primary
            else cls.union_value_type(primary.value_type, secondary.value_type),
            instance_source=instance_source,
            inherited=primary.inherited,
            physical=primary.physical,
        )

    @staticmethod
    def _union_min_count(primary: int | None, secondary: int | None) -> int | None:
        if primary is None:
            return secondary
        if secondary is None:
            return primary
        return min(primary, secondary)

    @staticmethod
    def _union_max_count(primary: int | float | None, secondary: int | float | None) -> int | float | None:
        if primary is None:
            return secondary
        if secondary is None:
            return primary
        output = max(primary, secondary)
        try:
            return int(output)
        except (OverflowError, ValueError):
            # The value is float('inf') or float('-inf')
            return output

    @staticmethod
    def union_value_type(
        primary: DataType | ConceptEntity | MultiValueTypeInfo | UnknownEntity,
        secondary: DataType | ConceptEntity | MultiValueTypeInfo | UnknownEntity,
    ) -> DataType | ConceptEntity | MultiValueTypeInfo | UnknownEntity:
        all_types: list[DataType | ConceptEntity] = []
        for type_ in (primary, secondary):
            if isinstance(type_, MultiValueTypeInfo):
                all_types.extend(type_.types)
            elif isinstance(type_, UnknownEntity):
                continue
            elif isinstance(type_, DataType | ConceptEntity):
                all_types.append(type_)
            else:
                raise NotImplementedError(f"Unsupported type: {type_}")

        ordered_types = list(dict.fromkeys(all_types))
        if len(ordered_types) == 0:
            return UnknownEntity()
        if len(ordered_types) == 1:
            return ordered_types[0]
        else:  # len(ordered_types) > 1:
            return MultiValueTypeInfo(types=ordered_types)



import warnings
from collections.abc import Mapping
from pathlib import Path
from typing import ClassVar, cast

import yaml
from openpyxl import load_workbook
from openpyxl.worksheet.worksheet import Worksheet
from pydantic import ValidationError

from cognite.neat._data_model.importers._base import DMSImporter
from cognite.neat._data_model.models.dms import (
    RequestSchema,
)
from cognite.neat._exceptions import DataModelImportException
from cognite.neat._issues import ModelSyntaxError
from cognite.neat._utils.text import humanize_collection
from cognite.neat._utils.useful_types import CellValueType, DataModelTableType
from cognite.neat._utils.validation import ValidationContext, as_json_path, humanize_validation_error

from .data_classes import MetadataValue, TableDMS
from .reader import DMSTableReader
from .source import SpreadsheetReadContext, TableSource


class DMSTableImporter(DMSImporter):
    """Imports DMS from a table structure.

    The tables can are expected to be a dictionary where the keys are the table names and the values
    are lists of dictionaries representing the rows in the table.
    """

    # We can safely cast as we know the validation_alias is always set to a str.
    REQUIRED_SHEETS = tuple(
        cast(str, field_.validation_alias) for field_ in TableDMS.model_fields.values() if field_.is_required()
    )
    REQUIRED_SHEET_MESSAGES: ClassVar[Mapping[str, str]] = {
        f"Missing required column: {sheet!r}": f"Missing required sheet: {sheet!r}" for sheet in REQUIRED_SHEETS
    }
    MetadataSheet = cast(str, TableDMS.model_fields["metadata"].validation_alias)

    def __init__(self, tables: DataModelTableType, source: TableSource | None = None) -> None:
        self._table = tables
        self._source = source or TableSource("Unknown")

    def to_data_model(self) -> RequestSchema:
        tables = self._read_tables()

        space, version = self._read_defaults(tables.metadata)
        reader = DMSTableReader(space, version, self._source)
        return reader.read_tables(tables)

    @classmethod
    def from_yaml(cls, yaml_file: Path) -> "DMSTableImporter":
        """Create a DMSTableImporter from a YAML file."""
        source = cls._display_name(yaml_file)
        return cls(yaml.safe_load(yaml_file.read_text()), TableSource(source.as_posix()))

    @classmethod
    def from_json(cls, json_file: Path) -> "DMSTableImporter":
        """Create a DMSTableImporter from a JSON file."""
        return cls.from_yaml(json_file)

    @classmethod
    def from_excel(cls, excel_file: Path) -> "DMSTableImporter":
        """Create a DMSTableImporter from an Excel file."""
        tables: DataModelTableType = {}
        source = TableSource(cls._display_name(excel_file).as_posix())
        workbook = load_workbook(excel_file, read_only=True, data_only=True, rich_text=False)
        try:
            for column_id, column_info in TableDMS.model_fields.items():
                sheet_name = cast(str, column_info.validation_alias)
                if sheet_name not in workbook.sheetnames:
                    continue
                required_columns = TableDMS.get_sheet_columns(column_id, column_info, column_type="required")
                sheet = workbook[sheet_name]
                context = SpreadsheetReadContext()
                table_rows = cls._read_rows(sheet, required_columns, context)
                tables[sheet_name] = table_rows
                source.table_read[sheet_name] = context
            return cls(tables, source)
        finally:
            workbook.close()

    def _read_tables(self) -> TableDMS:
        try:
            # Check tables, columns, data type and entity syntax.
            table = TableDMS.model_validate(self._table)
        except ValidationError as e:
            errors = self._create_error_messages(e)
            raise DataModelImportException(errors) from None
        return table

    def _create_error_messages(self, error: ValidationError) -> list[ModelSyntaxError]:
        errors: list[ModelSyntaxError] = []
        context = ValidationContext(
            humanize_location=self._location,
            field_name="column",
            missing_required_descriptor="missing",
        )
        seen: set[str] = set()

        for error_details in error.errors(include_input=True, include_url=False):
            message = humanize_validation_error(error_details, context)
            # Replace messages about missing required columns with missing required sheets.
            message = self.REQUIRED_SHEET_MESSAGES.get(message, message)
            if message in seen:
                # We treat all rows as the same, so we get duplicated errors for each row.
                continue

            seen.add(message)
            errors.append(ModelSyntaxError(message=message))

        return errors

    def _location(self, loc: tuple[str | int, ...]) -> str:
        if isinstance(loc[0], str) and len(loc) == 2:  # Sheet + row.
            # We skip the row as we treat all rows as the same. For example, if a required column is missing in one
            # row, it is missing in all rows.
            return f"{loc[0]} sheet"
        elif len(loc) == 3 and isinstance(loc[0], str) and isinstance(loc[1], int) and isinstance(loc[2], str):
            # This means there is something wrong in a specific cell.

            sheet = loc[0]
            row = loc[1]
            if self._source and sheet in self._source.table_read:
                context = self._source.table_read[sheet]
                row = context.adjusted_row_number(row) - 1

            return f"{sheet} sheet row {row + 1} column {loc[2]!r}"
        # This should be unreachable as the TableDMS model only has 2 levels.
        return as_json_path(loc)

    @staticmethod
    def _read_defaults(metadata: list[MetadataValue]) -> tuple[str, str]:
        """Reads the space and version from the metadata table."""
        default_space: str | None = None
        default_version: str | None = None
        missing = {"space", "version"}
        for meta in metadata:
            if meta.key == "space":
                default_space = str(meta.value)
                missing.remove("space")
            elif meta.key == "version":
                default_version = str(meta.value)
                missing.remove("version")
        if missing:
            error = ModelSyntaxError(message=f"In Metadata missing required values: {humanize_collection(missing)}")
            # If space or version is missing, we cannot continue parsing the model as these are used as defaults.
            raise DataModelImportException([error]) from None
        return str(default_space), str(default_version)

    @classmethod
    def _display_name(cls, filepath: Path) -> Path:
        """Get a display-friendly version of the file path."""
        cwd = Path.cwd()
        source = filepath
        if filepath.is_relative_to(cwd):
            source = filepath.relative_to(cwd)
        return source

    @classmethod
    def _read_rows(
        cls, sheet: Worksheet, required_columns: list[str], context: SpreadsheetReadContext
    ) -> list[dict[str, CellValueType]]:
        table_rows: list[dict[str, CellValueType]] = []
        # Metadata sheet is just a key-value pair of the first two columns.
        # For other sheets, we need to find the column header row first.
        columns: list[str] = [] if sheet.title != cls.MetadataSheet else required_columns
        # Ignore warnings about Data Validation extension not being supported in read-only mode.
        warnings.filterwarnings("ignore", category=UserWarning, message="Data Validation extension is not supported*")
        # The .iter_rows with values_only=True raises the warning `UserWarning: Data Validation extension is
        # not supported and will be removed`
        # which confuses the user and we do not depend on data validation here, so we suppress it.
        for row_no, row in enumerate(sheet.iter_rows(values_only=True)):
            if columns:
                # We have found the column header row, read the data rows.
                if all(cell is None for cell in row):
                    context.empty_rows.append(row_no)
                else:
                    record = dict(zip(columns, row, strict=False))
                    # MyPy complains as it thinks DataTableFormula | ArrayFormula could be cell values,
                    # but as we used values_only=True, this is not the case.
                    table_rows.append(record)  # type: ignore[arg-type]
            else:
                # Look for the column header row.
                row_values = [str(cell) for cell in row]
                if set(row_values).intersection(required_columns):
                    columns = row_values
                    context.header_row = row_no

        return table_rows



from dataclasses import dataclass

from cognite.neat._v0.core._issues.errors import _NEAT_ERRORS_BY_NAME, NeatError


@dataclass(unsafe_hash=True)
class PluginError(NeatError, ImportError):
    """No plugin of type '{plugin_type}' registered under name '{plugin_name}'"""

    plugin_name: str
    plugin_type: str


@dataclass(unsafe_hash=True)
class PluginLoadingError(PluginError):
    """Unable to load plugin of type '{plugin_type}' registered under name '{plugin_name}' due to: {exception}"""

    exception: str


@dataclass(unsafe_hash=True)
class PluginDuplicateError(PluginError):
    """Plugin of type '{plugin_type}' registered for under name '{plugin_name}' already exists"""

    ...


# Register the errors in the _NEAT_ERRORS_BY_NAME dictionary
_NEAT_ERRORS_BY_NAME.update(
    {
        "PluginError": PluginError,
        "PluginLoadingError": PluginLoadingError,
        "PluginDuplicateError": PluginDuplicateError,
    }
)



"""Validators checking for consistency issues in data model."""

from cognite.neat._data_model._constants import COGNITE_SPACES
from cognite.neat._data_model.rules.dms._base import DataModelRule
from cognite.neat._issues import Recommendation

BASE_CODE = "NEAT-DMS-CONSISTENCY"


class ViewSpaceVersionInconsistentWithDataModel(DataModelRule):
    """Validates that views have consistent space and version with the data model.

    ## What it does
    Validates that all views in the data model have the same space and version as the data model.

    ## Why is this bad?
    If views have different space or version than the data model, it may lead to more demanding development and
    maintenance efforts. The industry best practice is to keep views in the same space and version as the data model.

    ## Example
    If the data model is defined in space "my_space" version "v1", but a view is defined in the same spave but with
    version "v2", this requires additional attention during deployment and maintenance of the data model.
    """

    code = f"{BASE_CODE}-001"
    issue_type = Recommendation

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        if not self.validation_resources.merged_data_model.views:
            return recommendations

        space = self.validation_resources.merged_data_model.space
        version = self.validation_resources.merged_data_model.version

        for view_ref in self.validation_resources.merged_data_model.views:
            issue_description = ""

            if view_ref.space not in COGNITE_SPACES:
                # notify about inconsistent space
                if view_ref.space != space:
                    issue_description = f"space (view: {view_ref.space}, data model: {space})"

                # or version if spaces are same
                elif view_ref.version != version:
                    issue_description = f"version (view: {view_ref.version}, data model: {version})"

            if issue_description:
                recommendations.append(
                    Recommendation(
                        message=(f"View {view_ref!s} has inconsistent {issue_description} with the data model."),
                        fix="Update view version and/or space to match data model",
                        code=self.code,
                    )
                )

        return recommendations



import json
import warnings
from pathlib import Path
from typing import Literal, get_args

import yaml

from cognite.neat._v0.core._data_model._shared import VerifiedDataModel

from ._base import BaseExporter


class YAMLExporter(BaseExporter[VerifiedDataModel, str]):
    """Export data_model (Information, DMS or Domain) to YAML.

    Args:
        files: The number of files to output. Defaults to "single".
        output: The format to output the data_model. Defaults to "yaml".

    The following formats are available:

    - "single": A single YAML file will contain the entire data_model.

    .. note::

        YAML files are typically used for storing data_model when checked into version
        control systems, e.g., git-history.The advantage of using YAML files over
        Excel is that tools like git can show the differences between different
        versions of the data_model.

    """

    Files = Literal["single"]
    Format = Literal["yaml", "json"]

    file_option = get_args(Files)
    format_option = get_args(Format)

    def __init__(self, files: Files = "single", output: Format = "yaml"):
        if files not in self.file_option:
            raise ValueError(f"Invalid files: {files}. Valid options are {self.file_option}")
        if output not in self.format_option:
            raise ValueError(f"Invalid output: {output}. Valid options are {self.format_option}")
        self.files = files
        self.output = output

    @property
    def description(self) -> str:
        return "Export verified model to YAML."

    def export_to_file(self, data_model: VerifiedDataModel, filepath: Path) -> None:
        """Exports transformation data_model to YAML/JSON file(s)."""
        if self.files == "single":
            if filepath.suffix != f".{self.output}":
                warnings.warn(f"File extension is not .{self.output}, adding it to the file name", stacklevel=2)
                filepath = filepath.with_suffix(f".{self.output}")
            filepath.write_text(self.export(data_model), encoding=self._encoding, newline=self._new_line)
        else:
            raise NotImplementedError(f"Exporting to {self.files} files is not supported")

    def export(self, data_model: VerifiedDataModel) -> str:
        """Export data_model to YAML (or JSON) format.

        This will export the data_model to YAML format if the output is
        set to "yaml" and JSON format if the output is set.
        All None and Unset values are excluded from the output
        to keep the output clean, i.e., only the values the user
        has set.

        Args:
            data_model: The data_model to be exported.

        Returns:
            str: The data_model in YAML (or JSON) format.
        """
        # model_dump_json ensures that the output is in JSON format,
        # if we don't do this, we will get Enums and other types that are not serializable to YAML
        json_output = data_model.dump(mode="json", sort=True, exclude_none=True, exclude_unset=True)
        if self.output == "json":
            return json.dumps(json_output)
        elif self.output == "yaml":
            return yaml.safe_dump(json_output)
        else:
            raise ValueError(f"Invalid output: {self.output}. Valid options are {self.format_option}")



from ._data_model import DataModelImporterPlugin, DataModelTransformerPlugin
from ._manager import get_plugin_manager

__all__ = ["DataModelImporterPlugin", "DataModelTransformerPlugin", "get_plugin_manager"]



from abc import ABC
from typing import Annotated, Literal

from pydantic import Field, TypeAdapter

from cognite.neat._utils.useful_types import BaseModelObject

from ._types import Bool


class IndexDefinition(BaseModelObject, ABC):
    index_type: str
    properties: list[str] = Field(description="List of properties to define the index across.")


class BtreeIndex(IndexDefinition):
    index_type: Literal["btree"] = "btree"
    by_space: Bool | None = Field(default=None, description="Whether to make the index space-specific.")
    cursorable: Bool | None = Field(
        default=None, description="Whether the index can be used for cursor-based pagination."
    )


class InvertedIndex(IndexDefinition):
    index_type: Literal["inverted"] = "inverted"


Index = Annotated[BtreeIndex | InvertedIndex, Field(discriminator="index_type")]

IndexAdapter: TypeAdapter[Index] = TypeAdapter(Index)



from dataclasses import dataclass
from typing import Generic, TypeAlias, TypeVar

from cognite.neat._v0.core._data_model.models import (
    ConceptualDataModel,
    PhysicalDataModel,
)
from cognite.neat._v0.core._data_model.models._import_contexts import ImportContext
from cognite.neat._v0.core._data_model.models.conceptual._unverified import (
    UnverifiedConceptualDataModel,
)
from cognite.neat._v0.core._data_model.models.physical._unverified import (
    UnverifiedPhysicalDataModel,
)

VerifiedDataModel: TypeAlias = ConceptualDataModel | PhysicalDataModel

T_VerifiedDataModel = TypeVar("T_VerifiedDataModel", bound=VerifiedDataModel)
UnverifiedDataModel: TypeAlias = UnverifiedPhysicalDataModel | UnverifiedConceptualDataModel
T_UnverifiedDataModel = TypeVar("T_UnverifiedDataModel", bound=UnverifiedDataModel)


@dataclass
class ImportedDataModel(Generic[T_UnverifiedDataModel]):
    """This class is used to store results of data model import from a source prior to
    verification and validation.

    Attributes:
        unverified_data_model: The unverified data model.
        context: The context of the import, including warnings, errors and any other
                 relevant information.
    """

    unverified_data_model: T_UnverifiedDataModel | None
    context: ImportContext | None = None

    @classmethod
    def display_type_name(cls) -> str:
        return "UnverifiedModel"

    @property
    def display_name(self) -> str:
        if self.unverified_data_model is None:
            return "Failed to load data model"
        return self.unverified_data_model.display_name


ImportedUnverifiedDataModel: TypeAlias = (
    ImportedDataModel[UnverifiedPhysicalDataModel] | ImportedDataModel[UnverifiedConceptualDataModel]
)
T_ImportedUnverifiedDataModel = TypeVar("T_ImportedUnverifiedDataModel", bound=ImportedUnverifiedDataModel)

DataModel: TypeAlias = (
    ConceptualDataModel
    | PhysicalDataModel
    | ImportedDataModel[UnverifiedPhysicalDataModel]
    | ImportedDataModel[UnverifiedConceptualDataModel]
)
T_DataModel = TypeVar("T_DataModel", bound=DataModel)



from pathlib import Path
from typing import Any

from cognite.neat._v0.core._data_model.importers._base import BaseImporter
from cognite.neat._v0.core._data_model.transformers._base import VerifiedDataModelTransformer

from ._base import NeatPlugin


class DataModelImporterPlugin(NeatPlugin):
    """This class is used an interface for data model import plugins.
    Any plugin that is used for importing data models should inherit from this class.
    It is expected to implement the `configure` method which returns a configured importer.
    """

    def configure(self, io: str | Path | None, **kwargs: Any) -> BaseImporter:
        """Return a configure plugin for data model import.
        Args:
            io (str | Path | None): The input/output interface for the plugin.
            **kwargs (Any): Additional keyword arguments for configuration.
        Returns:
            BaseImporter: A configured instance of the BaseImporter.
        !!! note "Returns"
            The method must return an instance of `BaseImporter` or its subclasses
            meaning it must implement the `BaseImporter` interface.
        """

        raise NotImplementedError()


class DataModelTransformerPlugin(NeatPlugin):
    """This class is used as an interface for data model transformer plugins.
    Any plugin that is used for transforming data models should inherit from this class.
    It is expected to implement the `configure` method which returns a configured transformer.
    """

    def configure(self, **kwargs: Any) -> VerifiedDataModelTransformer:
        """Return a configure plugin for data model import.
        Args:
            **kwargs (Any): Additional keyword arguments for configuration.
        Returns:
            VerifiedDataModelTransformer: A configured instance of the VerifiedDataModelTransformer.
        !!! note "Returns"
            The method must return an instance of `VerifiedDataModelTransformer` or its subclasses
            meaning it must implement the `VerifiedDataModelTransformer` interface.
        """

        raise NotImplementedError()



from collections.abc import Callable
from datetime import datetime, timezone

from cognite.neat._data_model._analysis import ValidationResources
from cognite.neat._data_model._shared import OnSuccessIssuesChecker
from cognite.neat._data_model._snapshot import SchemaSnapshot
from cognite.neat._data_model.models.dms._limits import SchemaLimits
from cognite.neat._data_model.models.dms._schema import RequestSchema
from cognite.neat._data_model.rules.dms._base import DataModelRule
from cognite.neat._utils.auxiliary import get_concrete_subclasses
from cognite.neat._utils.useful_types import ModusOperandi


class DmsDataModelRulesOrchestrator(OnSuccessIssuesChecker):
    """DMS Data Model rules orchestrator, used to execute DMS data model rules on a single data model represented
    as RequestSchema."""

    def __init__(
        self,
        cdf_snapshot: SchemaSnapshot,
        limits: SchemaLimits,
        modus_operandi: ModusOperandi = "additive",
        can_run_validator: Callable[[str, type], bool] | None = None,
        enable_alpha_validators: bool = False,
    ) -> None:
        super().__init__()
        self._cdf_snapshot = cdf_snapshot
        self._limits = limits
        self._modus_operandi = modus_operandi
        self._can_run_validator = can_run_validator or (lambda code, issue_type: True)  # type: ignore
        self._has_run = False
        self._enable_alpha_validators = enable_alpha_validators

    def run(self, request_schema: RequestSchema) -> None:
        """Run quality assessment on the DMS data model."""

        validation_resources = self._gather_validation_resources(request_schema)

        # Initialize all validators
        validators: list[DataModelRule] = [
            validator(validation_resources) for validator in get_concrete_subclasses(DataModelRule)
        ]

        # Run validators
        for validator in validators:
            if validator.alpha and not self._enable_alpha_validators:
                continue
            if self._can_run_validator(validator.code, validator.issue_type):
                self._issues.extend(validator.validate())
                if validator.fixable:
                    self._pending_fixes.extend(validator.fix())

        self._has_run = True

    def copy(self) -> "DmsDataModelRulesOrchestrator":
        return DmsDataModelRulesOrchestrator(
            cdf_snapshot=self._cdf_snapshot,
            limits=self._limits,
            modus_operandi=self._modus_operandi,
            can_run_validator=self._can_run_validator,
            enable_alpha_validators=self._enable_alpha_validators,
        )

    def _gather_validation_resources(self, request_schema: RequestSchema) -> ValidationResources:
        # we do not want to modify the original request schema during validation
        copy = request_schema.model_copy(deep=True)
        local = SchemaSnapshot(
            data_model={request_schema.data_model.as_reference(): copy.data_model},
            views={view.as_reference(): view for view in copy.views},
            containers={container.as_reference(): container for container in copy.containers},
            spaces={space.as_reference(): space for space in copy.spaces},
            node_types={node_type: node_type for node_type in copy.node_types},
            timestamp=datetime.now(timezone.utc),
        )

        return ValidationResources(
            cdf=self._cdf_snapshot,
            local=local,
            limits=self._limits,
            modus_operandi=self._modus_operandi,
        )



"""Validators for connections in data model specifications."""

from dataclasses import dataclass

from cognite.neat._data_model.models.dms._data_types import DirectNodeRelation
from cognite.neat._data_model.models.dms._references import (
    ViewDirectReference,
    ViewReference,
)
from cognite.neat._data_model.models.dms._view_property import ViewCorePropertyRequest
from cognite.neat._data_model.rules.dms._base import DataModelRule
from cognite.neat._issues import ConsistencyError, Recommendation

BASE_CODE = "NEAT-DMS-CONNECTIONS"


class ConnectionValueTypeUnexisting(DataModelRule):
    """Validates that connection value types exist.

    ## What it does
    Validates that all connection value types defined in the data model exist.

    ## Why is this bad?
    If a connection value type does not exist, the data model cannot be deployed to CDF.
    This means that the connection will not be able to function.

    ## Example
    If view WindTurbine has a connection property windFarm with value type WindFarm, but WindFarm view is not defined,
    the data model cannot be deployed to CDF.
    """

    code = f"{BASE_CODE}-001"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        for (view, property_), value_type in self.validation_resources.connection_end_node_types.items():
            if value_type is None:
                continue

            if self.validation_resources.select_view(value_type) is not None:
                continue

            errors.append(
                ConsistencyError(
                    message=(
                        f"View {view!s} connection {property_!s} has value type {value_type!s} "
                        "which is not defined as a view in the data model neither exists in CDF."
                    ),
                    fix="Define necessary view",
                    code=self.code,
                )
            )

        return errors


class ConnectionValueTypeUndefined(DataModelRule):
    """Validates that connection value types are not None, i.e. undefined.

    ## What it does
    Validates that connections have explicitly defined value types (i.e., end connection node type).

    ## Why is this bad?
    If a connection value type is None (undefined), there is no type information about the end node of the connection.
    This yields an ambiguous data model definition, which may lead to issues during consumption of data from CDF.

    ## Example
    Consider a scenario where we have views WindTurbine,ArrayCable and Substation. Lets say WindTurbine has a connection
    `connectsTo` with value type None (undefined), then it is unclear what type of view the connection points to as
    both ArrayCable and Substation are valid targets for the connection.

    """

    code = f"{BASE_CODE}-002"
    issue_type = Recommendation

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        for (view, property_), value_type in self.validation_resources.connection_end_node_types.items():
            if not value_type:
                recommendations.append(
                    Recommendation(
                        message=(
                            f"View {view!s} connection {property_!s} is missing value type (end node type)."
                            " This yields ambiguous data model definition."
                        ),
                        fix="Define necessary value type",
                        code=self.code,
                    )
                )

        return recommendations


@dataclass
class ReverseConnectionContext:
    """Context for validating a bidirectional connection.
    This context holds all necessary references to validate the connection.

    Attributes:
        target_view_ref: Reference to the target view containing the reverse property.
        reverse_property: Identifier of the reverse property in the target view.
        through: Direct reference defining the direct property used in the reverse connection.
        source_view_ref: Reference to the source view containing the direct property, to which the reverse points.


    Example:
        View `WindTurbine` has property `windFarm` (direct) → points to View `WindFarm`
        View `WindFarm` has property `turbines` (reverse) → points to View `WindTurbine` through `WindTurbine.windFarm`
    """

    target_view_ref: ViewReference
    reverse_property: str
    through: ViewDirectReference
    source_view_ref: ViewReference


class ReverseConnectionSourceViewMissing(DataModelRule):
    """Validates that source view referenced in reverse connection exist.

    ## What it does
    Checks that the source view used to configure a reverse connection exists.

    ## Why is this bad?
    A reverse connection requires a corresponding direct connection in the source view.
    If the source view doesn't exist, the reverse connection is invalid.

    ## Example
    If WindFarm has a reverse property `turbines` through `WindTurbine.windFarm`,
    but WindTurbine view doesn't exist, the reverse connection cannot function.
    """

    code = f"{BASE_CODE}-REVERSE-001"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        for (target_view_ref, reverse_prop_name), (
            source_view_ref,
            through,
        ) in self.validation_resources.reverse_to_direct_mapping.items():
            through = self.validation_resources.normalize_through_reference(source_view_ref, through)
            source_view = self.validation_resources.select_view(source_view_ref, through.identifier)

            if not source_view:
                errors.append(
                    ConsistencyError(
                        message=(
                            f"Source view {source_view_ref!s} used to configure reverse connection "
                            f"'{reverse_prop_name}' in target view {target_view_ref!s} "
                            "does not exist in the data model or CDF."
                        ),
                        fix="Define the missing source view",
                        code=self.code,
                    )
                )

        return errors


class ReverseConnectionSourcePropertyMissing(DataModelRule):
    """Validates that source property referenced in reverse connections exist.

    ## What it does
    Checks that the direct connection property in the source view (used in the reverse connection's 'through')
    actually exists in the source view.

    ## Why is this bad?
    A reverse connection requires a corresponding direct connection property in the source view.
    If this property doesn't exist, the bidirectional connection is incomplete.

    ## Example
    If WindFarm has a reverse property `turbines` through `WindTurbine.windFarm`,
    but WindTurbine view doesn't have a `windFarm` property, the reverse connection is invalid.
    """

    code = f"{BASE_CODE}-REVERSE-002"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        for (target_view_ref, reverse_prop_name), (
            source_view_ref,
            through,
        ) in self.validation_resources.reverse_to_direct_mapping.items():
            through = self.validation_resources.normalize_through_reference(source_view_ref, through)
            source_view = self.validation_resources.select_view(source_view_ref, through.identifier)

            if not source_view:
                continue  # Handled by ReverseConnectionSourceViewMissing

            # critical to expand view properties to include inherited ones as otherwise we might miss the property
            if not (source_view_expanded := self.validation_resources.expand_view_properties(source_view_ref)):
                raise RuntimeError(f"{type(self).__name__}: View {source_view_ref!s} not found. This is a bug in NEAT.")

            if not source_view_expanded.properties or through.identifier not in source_view_expanded.properties:
                errors.append(
                    ConsistencyError(
                        message=(
                            f"Source view {source_view_ref!s} is missing property '{through.identifier}' "
                            f"which is required to configure the reverse connection "
                            f"'{reverse_prop_name}' in target view {target_view_ref!s}."
                        ),
                        fix="Add the missing property to the source view",
                        code=self.code,
                    )
                )

        return errors


class ReverseConnectionSourcePropertyWrongType(DataModelRule):
    """Validates that source property for the reverse connections is a direct relation.

    ## What it does
    Checks that the property referenced in a reverse connection's 'through' clause
    is actually a direct connection property (not a primitive or other type).

    ## Why is this bad?
    Reverse connections can only work with direct connection properties.
    Using other property types breaks the bidirectional relationship.

    ## Example
    If WindFarm has a reverse property `turbines` through `WindTurbine.name`,
    but `name` is a Text property (not a direct connection), the reverse connection is invalid.
    """

    code = f"{BASE_CODE}-REVERSE-003"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        for (target_view_ref, reverse_prop_name), (
            source_view_ref,
            through,
        ) in self.validation_resources.reverse_to_direct_mapping.items():
            through = self.validation_resources.normalize_through_reference(source_view_ref, through)
            source_view = self.validation_resources.select_view(source_view_ref, through.identifier)

            if not source_view:
                continue  # Handled by ReverseConnectionSourceViewMissing

            if not (source_view_expanded := self.validation_resources.expand_view_properties(source_view_ref)):
                raise RuntimeError(f"{type(self).__name__}: View {source_view_ref!s} not found. This is a bug in NEAT.")

            if not source_view_expanded.properties or through.identifier not in source_view_expanded.properties:
                continue  # Handled by ReverseConnectionSourcePropertyMissing

            source_property = source_view_expanded.properties[through.identifier]

            if not isinstance(source_property, ViewCorePropertyRequest):
                errors.append(
                    ConsistencyError(
                        message=(
                            f"Source view {source_view_ref!s} property '{through.identifier}' "
                            f"used for configuring the reverse connection '{reverse_prop_name}' "
                            f"in target view {target_view_ref!s} is not a direct connection property."
                        ),
                        fix="Update view property to be a direct connection property",
                        code=self.code,
                    )
                )

        return errors


class ReverseConnectionContainerMissing(DataModelRule):
    """Validates that the container referenced by the reverse connection source properties exist.

    ## What it does
    Checks that the container holding the direct connection property (used in reverse connection) exists.

    ## Why is this bad?
    The direct connection property must be stored in a container.
    If the container doesn't exist, the connection cannot be persisted.

    ## Example
    If WindTurbine.windFarm maps to container `WindTurbine`, but this container doesn't exist,
    the reverse connection from WindFarm cannot function.
    """

    code = f"{BASE_CODE}-REVERSE-004"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        for resolved in self.validation_resources.resolved_reverse_direct_relations:
            if resolved.container is None:
                errors.append(
                    ConsistencyError(
                        message=(
                            f"Container {resolved.container_ref!s} is missing in both the data model and CDF. "
                            f"This container is required by view {resolved.direct_view_ref!s}"
                            f" property '{resolved.through_property_id}', "
                            f"which configures the reverse connection '{resolved.reverse_property_id}'"
                            f" in target view {resolved.reverse_view_ref!s}."
                        ),
                        fix="Define the missing container",
                        code=self.code,
                    )
                )

        return errors


class ReverseConnectionContainerPropertyMissing(DataModelRule):
    """Validates that container property referenced by the reverse connections exists.

    ## What it does
    Checks that the property in the container (mapped from the view's direct connection property)
    actually exists in the container.

    ## Why is this bad?
    The view property must map to an actual container property for data persistence.
    If the container property doesn't exist, data cannot be stored.

    ## Example
    If WindTurbine.windFarm maps to container property `WindTurbine.windFarm`,
    but this container property doesn't exist, the connection cannot be stored.
    """

    code = f"{BASE_CODE}-REVERSE-005"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        for resolved in self.validation_resources.resolved_reverse_direct_relations:
            # Container must exist but property is missing
            if resolved.container is not None and resolved.container_property is None:
                errors.append(
                    ConsistencyError(
                        message=(
                            f"Container {resolved.container_ref!s} is missing "
                            f"property '{resolved.container_property_id}'. "
                            f"This property is required by the source view {resolved.direct_view_ref!s}"
                            f" property '{resolved.through_property_id}', "
                            f"which configures the reverse connection '{resolved.reverse_property_id}' "
                            f"in target view {resolved.reverse_view_ref!s}."
                        ),
                        fix="Add the missing property to the container",
                        code=self.code,
                    )
                )

        return errors


class ReverseConnectionContainerPropertyWrongType(DataModelRule):
    """Validates that the container property used in reverse connection is the direct relations.

    ## What it does
    Checks that the container property (mapped from view's direct connection property)
    has type DirectNodeRelation.

    ## Why is this bad?
    Container properties backing connection view properties must be DirectNodeRelation type.
    Other types cannot represent connections in the underlying storage.

    ## Example
    If WindTurbine.windFarm maps to container property with type Text instead of DirectNodeRelation,
    the connection cannot be stored correctly.
    """

    code = f"{BASE_CODE}-REVERSE-006"
    issue_type = ConsistencyError

    def validate(self) -> list[ConsistencyError]:
        errors: list[ConsistencyError] = []

        for resolved in self.validation_resources.resolved_reverse_direct_relations:
            if not resolved.container_property:
                continue  # Handled by ReverseConnectionContainerPropertyMissing

            if not isinstance(resolved.container_property.type, DirectNodeRelation):
                errors.append(
                    ConsistencyError(
                        message=(
                            f"Container property '{resolved.container_property_id}' "
                            f"in container {resolved.container_ref!s} "
                            f"must be a direct connection, but found type '{resolved.container_property.type!s}'. "
                            f"This property is used by source view {resolved.direct_view_ref!s} "
                            f"property '{resolved.through_property_id}' "
                            f"to configure reverse connection '{resolved.reverse_property_id}' "
                            f"in target view {resolved.reverse_view_ref!s}."
                        ),
                        fix="Change container property type to be a direct connection",
                        code=self.code,
                    )
                )

        return errors


class ReverseConnectionTargetMissing(DataModelRule):
    """Validates that the direct connection in reverse connection pair have target views specified.

    ## What it does
    Checks whether the direct connection property (referenced by reverse connection) has a value type.

    ## Why is this bad?
    While CDF allows value type None as a SEARCH hack for multi-value relations,
    it's better to explicitly specify the target view for clarity and maintainability.

    ## Example
    If WindTurbine.windFarm has value type None instead of WindFarm,
    this validator recommends specifying WindFarm explicitly.
    """

    code = f"{BASE_CODE}-REVERSE-007"
    issue_type = Recommendation

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        for resolved in self.validation_resources.resolved_reverse_direct_relations:
            if not resolved.direct_property.source:
                recommendations.append(
                    Recommendation(
                        message=(
                            f"Source view {resolved.direct_view_ref!s} property '{resolved.through_property_id}' "
                            f"has no target view specified (value type is None). "
                            f"This property is used for reverse connection '{resolved.reverse_property_id}' "
                            f"in target view {resolved.reverse_view_ref!s}. "
                            f"While this works as a hack for multi-value relations in CDF Search, "
                            f"it's recommended to explicitly define the target view as {resolved.reverse_view_ref!s}."
                        ),
                        fix="Set the property's value type to the target view for better clarity",
                        code=self.code,
                    )
                )

        return recommendations


class ReverseConnectionPointsToAncestor(DataModelRule):
    """Validates that direct connections point to specific views rather than ancestors.

    ## What it does
    Checks whether the direct connection property points to an ancestor of the expected target view
    and recommends pointing to the specific target instead.

    ## Why is this bad?
    While technically valid, pointing to ancestors can be confusing and may lead to mistakes.
    It's clearer to point to the specific target view.

    ## Example
    If WindFarm.turbines expects WindTurbine.windFarm to point to WindFarm,
    but it points to Asset (ancestor of WindFarm), this validator recommends the change.
    """

    code = f"{BASE_CODE}-REVERSE-008"
    issue_type = Recommendation

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        for resolved in self.validation_resources.resolved_reverse_direct_relations:
            actual_target_view = resolved.direct_property.source

            if not actual_target_view:
                continue  # Handled by ReverseConnectionTargetMissing

            if self.validation_resources.is_ancestor(resolved.reverse_view_ref, actual_target_view):
                recommendations.append(
                    Recommendation(
                        message=(
                            f"The direct connection property '{resolved.through_property_id}' "
                            f"in view {resolved.direct_view_ref!s} "
                            f"configures the reverse connection '{resolved.reverse_property_id}' "
                            f"in {resolved.reverse_view_ref!s}. "
                            f"Therefore, it is expected that '{resolved.through_property_id}' "
                            f"points to {resolved.reverse_view_ref!s}. "
                            f"However, it currently points to {actual_target_view!s}, which is an ancestor of "
                            f"{resolved.reverse_view_ref!s}. "
                            "While this will allow for model to be valid, it can be a source of confusion and mistakes."
                        ),
                        fix="Update the direct connection property to point to the target view instead of its ancestor",
                        code=self.code,
                    )
                )

        return recommendations


class ReverseConnectionTargetMismatch(DataModelRule):
    """Validates that direct connections point to the correct target views.

    ## What it does
    Checks that the direct connection property points to the expected target view
    (the view containing the reverse connection).

    ## Why is this bad?
    The reverse connection expects a bidirectional relationship.
    If the direct connection points to a different view, the relationship is broken.

    ## Example
    If WindFarm.turbines is a reverse through WindTurbine.windFarm,
    but WindTurbine.windFarm points to SolarFarm instead of WindFarm, the connection is invalid.
    """

    code = f"{BASE_CODE}-REVERSE-009"
    issue_type = Recommendation

    def validate(self) -> list[Recommendation]:
        recommendations: list[Recommendation] = []

        for resolved in self.validation_resources.resolved_reverse_direct_relations:
            actual_target_view = resolved.direct_property.source

            if not actual_target_view:
                continue  # Handled by ReverseConnectionTargetMissing

            if self.validation_resources.is_ancestor(resolved.reverse_view_ref, actual_target_view):
                continue  # Handled by ReverseConnectionPointsToAncestor

            if actual_target_view != resolved.reverse_view_ref:
                recommendations.append(
                    Recommendation(
                        message=(
                            f"The reverse connection '{resolved.reverse_property_id}' "
                            f"in view {resolved.reverse_view_ref!s} "
                            f"expects its corresponding direct connection in view {resolved.direct_view_ref!s} "
                            f"(property '{resolved.through_property_id}') "
                            f"to point back to {resolved.reverse_view_ref!s}, "
                            f"but it actually points to {actual_target_view!s}."
                        ),
                        fix="Update the direct connection property to point back to the correct target view",
                        code=self.code,
                    )
                )

        return recommendations



import re
from collections.abc import Collection
from typing import Any

NEWLINE = "\n"
TAB = "\t"


def humanize_collection(collection: Collection[Any], /, *, sort: bool = True, bind_word: str = "and") -> str:
    """Convert a collection of items to a human-readable string.

    Args:
        collection: The collection of items to convert.
        sort: Whether to sort the collection before converting. Default is True.
        bind_word: The word to use to bind the last two items. Default is "and".

    Returns:
        A human-readable string of the collection.

    Examples:
        >>> humanize_collection(["b", "c", "a"])
        'a, b and c'
        >>> humanize_collection(["b", "c", "a"], sort=False)
        'b, c and a'
        >>> humanize_collection(["a", "b"])
        'a and b'
        >>> humanize_collection(["a"])
        'a'
        >>> humanize_collection([])
        ''

    """
    if not collection:
        return ""
    elif len(collection) == 1:
        return str(next(iter(collection)))

    strings = (str(item) for item in collection)
    if sort:
        sequence = sorted(strings)
    else:
        sequence = list(strings)

    return f"{', '.join(sequence[:-1])} {bind_word} {sequence[-1]}"


def title_case(s: str) -> str:
    """Convert a string to title case, handling underscores and hyphens.

    Args:
        s: The string to convert.
    Returns:
        The title-cased string.
    Examples:
        >>> title_case("hello world")
        'Hello World'
        >>> title_case("hello_world")
        'Hello World'
        >>> title_case("hello-world")
        'Hello World'
        >>> title_case("hello_world-and-universe")
        'Hello World And Universe'
        >>> title_case("HELLO WORLD")
        'Hello World'
    """
    return " ".join(word.capitalize() for word in s.replace("_", " ").replace("-", " ").split())


def split_on_capitals(text: str) -> list[str]:
    """Split a string at capital letters."""
    return re.findall(r"[A-Z][a-z]*", text)



import warnings
from pathlib import Path
from typing import Any

from cognite.neat._v0.core._issues._base import IssueList
from cognite.neat._v0.core._utils.reader._base import NeatReader
from cognite.neat._v0.plugins import DataModelImporterPlugin, get_plugin_manager
from cognite.neat._v0.session._experimental import ExperimentalFlags

from ._state import SessionState
from .exceptions import session_class_wrapper


@session_class_wrapper
class PluginAPI:
    """Read from a data source into NeatSession graph store."""

    def __init__(self, state: SessionState) -> None:
        self._state = state

        self.data_model = DataModelPlugins(self._state)


@session_class_wrapper
class DataModelPlugins:
    """Read from a data source into NeatSession graph store."""

    def __init__(self, state: SessionState) -> None:
        self._state = state

    def read(self, name: str, io: str | Path | None = None, **kwargs: Any) -> IssueList:
        """Provides access to the external plugins for data model importing.

        Args:
            name (str): The name of format (e.g. Excel) plugin is handling.
            io (str | Path | None): The input/output interface for the plugin.
            **kwargs (Any): Additional keyword arguments for the plugin.

        !!! note "kwargs"
            Users must consult the documentation of the plugin to understand
            what keyword arguments are supported.
        """
        warnings.filterwarnings("default")
        ExperimentalFlags.plugin.warn()

        # Some plugins may not support the io argument
        if io:
            reader = NeatReader.create(io)
            path = reader.materialize_path()
        else:
            path = None

        self._state._raise_exception_if_condition_not_met(
            "Data Model Read",
            empty_data_model_store_required=True,
        )

        plugin_manager = get_plugin_manager()
        plugin = plugin_manager.get(name, DataModelImporterPlugin)

        print(
            f"You are using an external plugin {plugin.__name__}, which is not developed by the NEAT team."
            "\nUse it at your own risk."
        )

        return self._state.data_model_import(plugin().configure(io=path, **kwargs))



from pydantic import HttpUrl, RootModel, ValidationError

from cognite.neat._utils.auxiliary import local_import


class URI(RootModel[str]):
    def __init__(self, value: str):
        try:
            # Use Pydantic's AnyUrl to validate the URI
            _ = HttpUrl(value)
        except ValidationError as e:
            raise ValueError(f"Invalid URI: {value}") from e
        super().__init__(value)

    def __str__(self) -> str:
        return self.root

    def __repr__(self) -> str:
        return f"URI({self.root!r})"

    def as_rdflib_uriref(self):  # type: ignore[no-untyped-def]
        # rdflib is an optional dependency, so import here
        local_import("rdflib", "rdflib")
        from rdflib import URIRef

        return URIRef(self.root)


class NameSpace(RootModel[str]):
    def __init__(self, value: str):
        try:
            # Use Pydantic's AnyUrl to validate the URI
            _ = HttpUrl(value)
        except ValidationError as e:
            raise ValueError(f"Invalid Namespace: {value}") from e
        super().__init__(value)

    def __str__(self) -> str:
        return self.root

    def __repr__(self) -> str:
        return f"NameSpace({self.root!r})"

    def term(self, name: str) -> URI:
        # need to handle slices explicitly because of __getitem__ override
        return URI(self.root + (name if isinstance(name, str) else ""))

    def __getitem__(self, key: str) -> URI:  # type: ignore[override]
        return self.term(key)

    def __getattr__(self, name: str) -> URI:
        if name.startswith("__"):  # ignore any special Python names!
            raise AttributeError
        return self.term(name)

    def as_rdflib_namespace(self):  # type: ignore[no-untyped-def]
        # rdflib is an optional dependency, so import here
        local_import("rdflib", "rdflib")
        from rdflib import Namespace

        return Namespace(self.root)



from abc import ABC

from pydantic import Field, field_validator

from cognite.neat._data_model.models.dms._references import SpaceReference
from cognite.neat._utils.text import humanize_collection

from ._base import APIResource, Resource, WriteableResource
from ._constants import FORBIDDEN_SPACES, SPACE_FORMAT_PATTERN


class Space(Resource, APIResource[SpaceReference], ABC):
    space: str = Field(
        description="The Space identifier (id).",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    name: str | None = Field(None, description="Name of the space.", max_length=255)
    description: str | None = Field(None, description="The description of the space.", max_length=1024)

    @field_validator("space")
    def check_forbidden_space_value(cls, val: str) -> str:
        """Check the space name not present in forbidden set"""
        if val in FORBIDDEN_SPACES:
            raise ValueError(f"{val!r} is a reserved space. Reserved Spaces: {humanize_collection(FORBIDDEN_SPACES)}")
        return val

    def as_reference(self) -> SpaceReference:
        return SpaceReference(space=self.space)


class SpaceRequest(Space): ...


class SpaceResponse(Space, WriteableResource[Space]):
    created_time: int = Field(
        description="When the space was created. The number of milliseconds since 00:00:00 Thursday, 1 January 1970, "
        "Coordinated Universal Time (UTC), minus leap seconds."
    )
    last_updated_time: int = Field(
        description="When the space was last updated. The number of milliseconds since 00:00:00 Thursday, "
        "1 January 1970, Coordinated Universal Time (UTC), minus leap seconds."
    )
    is_global: bool = Field(description="Whether the space is a global space.")

    def as_request(self) -> "SpaceRequest":
        return SpaceRequest.model_validate(self.model_dump(by_alias=True))
